#!/usr/bin/env python3
# Copyright (c) 2011-2025, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# wrapper for perf for using named events and events with additional MSRs.
# syntax is like perf, except Intel events are listed and can be specified
#
# or library for other python program to convert intel event names to
# perf/raw format
#
# Features:
# - map intel events to raw perf events
# - enable disable workarounds for specific events
# - resolve uncore events
# - handle offcore event on older kernels
# For the later must run as root and only as a single instance per machine
# Normal events (mainly not OFFCORE) can be handled unprivileged
# For events you can specify additional intel names from the list
#
# env variables:
# PERF=... perf binary to use (default "perf")
# EVENTMAP=eventmap
# EVENTMAP2=eventmap
# EVENTMAP3=eventmap
# OFFCORE=eventmap
# UNCORE=eventmap
# UNCORE2=eventmap
# eventmap is a path name to a json file. can contain wildcards.
# When eventmap is not specified, look in ~/.cache/pmu-events/
# The eventmap is automatically downloaded there
# eventmap can be also a CPU identifer (GenuineIntel-FAMILY-MODEL, like GenuineIntel-06-37)
# (Note that the numbers are in upper case hex)
#
# TOPOLOGY=topologyfile
# topologyfile is a dump of the sysfs of another system (find /sys > file)
# Needed for uncore units. This is useful to generate perf command lines for other systems.
#
# OCVERBOSE=1 print which files are opened
#
# Special arguments:
# --no-period   Never add a period
# --print       only print
# --force-download Force event list download
# --experimental Support experimental events
# --noexplode  Don't list all sub pmus for uncore events. Rely on perf stat to merge.
from __future__ import print_function
import sys
import os
import subprocess
import json
import re
import copy
import textwrap
if sys.version_info.major == 2:
    from pipes import quote
else:
    from shlex import quote
import itertools
import glob
from pmudef import EVENTSEL_ANY, EVENTSEL_INV, EVMASK, extra_flags, EVENTSEL_UMASK2
if sys.version_info.major == 3:
    import typing # noqa
    from typing import Set, List, Dict, Any, Tuple, DefaultDict # noqa

import msr as msrmod
import latego
import event_download

force_download = False
experimental = False
ocverbose = os.getenv("OCVERBOSE") is not None

exists_cache = dict() # type: Dict[str,bool]

emap_list = [] # type: List[EmapNativeJSON]

topology = None # type: None | Set[str]

def file_exists(s):
    if s in exists_cache:
        return exists_cache[s]
    global topology
    if topology is None:
        top = os.getenv("TOPOLOGY")
        topology = set()
        if top:
            try:
                topology = {x.strip()
                             .replace("/sys/devices", "/sys/bus/event_source/devices")
                            for x in open(top).readlines()}
            except OSError:
                print("Cannot open topology", top, file=sys.stderr)
    if s in topology:
        return True
    found = os.path.exists(s)
    exists_cache[s] = found
    return found

def has_format(s, pmu):
    return file_exists("/sys/bus/event_source/devices/%s/format/%s" % (pmu, s))

def has_format_any(f, pmu):
    return has_format(f, pmu) or has_format(f, pmu + "_0")

cached_vals = dict()

def cached_read(fn):
    if fn not in cached_vals:
        try:
            with open(fn) as f:
                cached_vals[fn] = f.read()
        except (OSError, IOError):
            cached_vals[fn] = "?"
    return cached_vals[fn]

warned = set()

def warn_once(s):
    if s not in warned:
        print(s, file=sys.stderr)
        warned.add(s)

class PerfVersion(object):
    def __init__(self):
        minor = 0
        perf = os.getenv("PERF")
        if not perf:
            perf = "perf"
        try:
            version = subprocess.Popen([perf, "--version"],
                    stdout=subprocess.PIPE).communicate()[0]
        except OSError:
            print("Cannot run", perf)
            version = ""
        if not isinstance(version, str):
            version = version.decode('utf-8')
        m = re.match(r"perf version (\d+)\.(\d+)\.", version)
        version = 412 # assume that no match is new enough
        if m:
            major = int(m.group(1))
            minor = int(m.group(2))
            version = major * 100 + minor

        pmu = "cpu_core" if os.path.exists("/sys/bus/event_source/devices/cpu_core") else "cpu"

        self.direct = os.getenv("DIRECT_MSR") or version < 400
        self.offcore = has_format("offcore_rsp", pmu) and not self.direct
        self.ldlat = has_format("ldlat", pmu) and not self.direct
        self.has_name = version >= 304
        self.has_uncore_expansion = version >= 412

version = PerfVersion()

class MSR(object):
    def __init__(self):
        self.reg = {}

    def writemsr(self, msrnum, val, print_only = False):
        print("msr %x = %x" % (msrnum, val, ))
        if print_only:
            return
        msrmod.writemsr(msrnum, val)

    def checked_writemsr(self, msr, val, print_only = False):
        if msr in self.reg:
            sys.exit("Multiple events use same register")
        self.writemsr(msr, val, print_only)
        self.reg[msr] = 1

qual_map = (
    ("amt1", "any=1", EVENTSEL_ANY, ""),
    ("percore", "percore=1", 0, ""),
    ("perf_metrics", "", 0, ""),
    ("i1", "inv=1", EVENTSEL_INV, ""),
    ("e1", "edge=1", 0, ""),
    ("e0", "edge=0", 0, ""),
    ("tx", "in_tx=1", 0, ""),
    ("sup", "", 0, "k"),
    ("usr=yes", "", 0, "u"),
    ("usr=no", "", 0, "k"),
    ("os=yes", "", 0, "k"),
    ("os=no", "", 0, "u"),
    ("anythr=yes", "any=1", 0, ""),
    ("anythr=no", "any=0", 0, ""),
    ("pdir", "", 0, "ppp"),
    ("precise=yes", "", 0, "pp"),
    ("cp", "in_tx_cp=1", 0, ""))

number = "(0x[0-9a-fA-F]+|[0-9]+)"

qualval_map = (
    (r"event_select=" + number, "event=%#x", 0),
    (r"u" + number, "umask=%#x", 0),
    (r"c(?:mask=)?" + number, "cmask=%d", 24),
    (r"e(?:dge=)?" + number, "edge=%d", 18),
    (r"ocr_msr_val=" + number, "config1=%#x", 0),
    (r"(?:sa|sample-after|period)=" + number, "period=%d", 0))

uncore_map = (
    (r'e(\d)', 'edge='),
    (r't=(\d+)', "thresh="),
    (r'match=(0x[0-9a-fA-F]+)', "filter_occ="),
    (r'filter1=(0x[0-9a-fA-F]+)', "config1=", 32),
    (r"nc=(\d+)", "filter_nc="),
    (r'filter=(0x[0-9a-fA-F]+)', "config1="),
    (r'one_unit', '', ),
    (r"u" + number, "umask="),
    (r"opc=?" + number, "filter_opc="),
    (r"tid=?" + number, "filter_tid="),
    (r"state=?" + number, "filter_state="))

uncore_map_thresh = (
    (r"c(?:mask=)?" + number, "thresh="),)

uncore_map_cmask = (
    (r"c(?:mask=)?" + number, "cmask="),)

# newe gets modified
def convert_extra(extra, val, newe):
    nextra = ""
    while extra:
        if extra[0] == ":":
            extra = extra[1:]
            continue
        found = False
        for j in qualval_map:
            m = re.match(j[0], extra, re.I)
            if m:
                if j[2]:
                    val |= int(m.group(1), 0) << j[2]
                newe.append(j[1] % (int(m.group(1), 0)))
                extra = extra[len(m.group(0)):]
                found = True
                break
        if found:
            continue
        found = False
        for j in qual_map:
            if extra.lower().startswith(j[0]):
                val |= j[2]
                newe.append(j[1])
                extra = extra[len(j[0]):]
                nextra += j[3]
                found = True
                break
        if found:
            continue
        if not extra:
            break
        if extra[0] in perf_qual + "p":
            nextra += extra[0]
            extra = extra[1:]
            continue
        print("bad event qualifier", extra, file=sys.stderr)
        break
    return nextra, val

def gen_name(n, sup):
    n = n.replace(".", "_").replace(":", "_").replace("=", "_")
    if sup:
        n += "_k"
    return n

class Event(object):
    def __init__(self, name, val, desc):
        self.val = val
        self.name = name
        self.extra = ""
        self.userextra = ""
        self.msr = 0
        self.msrval = 0
        self.desc = desc
        self.precise = 0
        self.collectpebs = 0
        self.newextra = ""
        self.overflow = None
        self.errata = None
        self.counter = ""
        self.period = 0
        self.pname = None

    # XXX return with pmu to be consistent with Uncore and fix callers
    def output_newstyle(self, extra="", noname=False, period=False, name="", noexplode=False):
        """Format an perf event for output and return as perf event string.
           Always uses new style (cpu/.../)."""
        val = self.val
        if extra:
            extra = self.newextra + "," + extra
        else:
            extra = self.newextra
        if self.pname:
            e = self.pname
        else:
            e = "event=0x%x,umask=0x%x" % (val & 0xff,
                                           ((val >> 8) & 0xff) | (((val >> 40) & 0xff) << 8))
        e += extra
        if version.has_name:
            if name:
                e += ",name=" + name
            elif not noname:
                e += ",name=%s" % (gen_name(self.name, "sup" in (self.extra + extra)))
        if period and self.period and ",period=" not in e:
            e += ",period=%d" % self.period
        return e

    def output(self, use_raw=False, flags="", noname=False, period=False, name="", noexplode=False):
        """Format an event for output and return as perf event string.
           use_raw when true return old style perf string (rXXX).
           Otherwise chose between old and new style based on the
           capabilities of the installed perf executable.
           flags when set add perf flags (e.g. u for user, p for pebs)."""
        val = self.val
        newe = []
        extra = "".join(sorted(merge_extra(extra_set(self.extra), extra_set(flags))))
        extra, val = convert_extra(":" + extra, val, newe)
        if version.direct or use_raw:
            if self.pname:
                ename = self.pname
            else:
                ename = "r%x" % (val,)
            if extra:
                ename += ":" + extra
            # XXX should error for extras that don't fit into raw
        else:
            p = self.output_newstyle(extra=",".join(newe),
                    noname=noname, period=period, name=name)
            ename = "%s/%s/" % (self.pmu, p) + extra
        return ename

    def filter_qual(self):
        if cached_read("/sys/bus/event_source/devices/%s/format/umask") == "config:8-15":
            self.val &= ~EVENTSEL_UMASK2

        def check_qual(q):
            if q == "":
                return True
            if "=" in q:
                q, _ = q.split("=")
            if has_format_any(q, self.pmu):
                return True
            warn_once("%s: format %s not supported. Filtering out" % (self. pmu, q))
            return False
        self.newextra = ",".join(filter(check_qual, self.newextra.split(",")))

box_to_perf = {
    "cbo": "cbox",
    "qpi_ll": "qpi",
    "sbo": "sbox",
}

def box_exists(box):
    return file_exists("/sys/bus/event_source/devices/uncore_%s" % (box))

def int_or_zero(row, name):
    if name in row:
        if row[name] == 'False':
            return 0
        if row[name] == 'True':
            return 1
        return int(row[name])
    return 0

uncore_units = {
    "imph-u": "arb",
    "kti ll": "upi",
    "m3kti": "m3upi",
    "upi ll": "upi",
}

def convert_uncore(flags, extra_map):
    o = ""
    while flags:
        for j in uncore_map + extra_map:
            if flags[0] == ",":
                flags = flags[1:]
            match, repl = j[0], j[1]
            m = re.match(match, flags, re.I)
            if m:
                if repl == "":
                    pass
                elif len(j) > 2:
                    o += "," + repl + ("%#x" % (int(m.group(1), 0) << j[2]))
                else:
                    o += "," + repl + m.group(1)
                flags = flags[m.end():]
            if flags == "":
                break
            if flags[0:1] == ":":
                flags = flags[1:]
        else:
            if flags != "":
                if len(extra_map) > 0:
                    print("Uncore cannot parse", flags, file=sys.stderr)
                break
    return o

class UncoreEvent(object):
    def __init__(self, name, row):
        self.name = name
        e = self
        if 'PublicDescription' in row:
            e.desc = row['PublicDescription'].strip()
        elif 'BriefDescription' in row:
            e.desc = row['BriefDescription'].strip()
        else:
            e.desc = row['Description'].strip()
        e.code = int(row['EventCode'], 16)
        if 'Internal' in row and int(row['Internal']) != 0:
            e.code |= int(row['Internal']) << 21
        e.umask = int(row['UMask'], 16)
        e.cmask = int_or_zero(row, 'CounterMask')
        e.inv = int_or_zero(row, 'Invert')
        e.edge = int_or_zero(row, 'EdgeDetect')
        e.unit = row['Unit'].lower()
        if e.unit in uncore_units:
            e.unit = uncore_units[e.unit]
        if e.unit == "ncu":
            e.unit = "clock" if box_exists("clock") else "cbox"
            e.umask = 0
            e.code = 0xff
        # xxx subctr
        if e.unit in box_to_perf:
            e.unit = box_to_perf[e.unit]
        e.msr = None
        e.overflow = 0
        e.counter = "1"  # dummy for toplev
        e.newextra = ""
        if 'Errata' in row:
            e.errata = row['Errata']
        else:
            e.errata = None
        self.extra = ''
        self.userextra = ''

    #  {
    # "Unit": "CBO",
    # "EventCode": "0x22",
    # "UMask": "0x21",
    # "EventName": "UNC_CBO_XSNP_RESPONSE.MISS_EXTERNAL",
    # "Description": "An external snoop misses in some processor core.",
    # "Counter": "0,1",
    # "CounterMask": "0",
    # "Invert": "0",
    # "EdgeDetect": "0"
    # },
    # XXX cannot separate sockets
    # extra: perf flags
    # flags: emon flags
    def output_newstyle(self, newextra="", noname=False, period=False, name="", flags="", noexplode=False):
        e = self
        o = "/event=%#x" % e.code
        if e.umask:
            o += ",umask=%#x" % e.umask
        if e.cmask:
            o += ",cmask=%#x" % e.cmask
        if e.edge:
            o += ",edge=1"
        if e.inv:
            o += ",inv=1"

        if e.newextra:
            if flags:
                flags += ","
            flags += e.newextra

        one_unit = "one_unit" in flags

        if has_format_any("cmask", "uncore_" + e.unit):
            extra_map = uncore_map_cmask
        else:
            extra_map = uncore_map_thresh

        o += convert_uncore(flags, extra_map)

        # xxx subctr, occ_sel, filters
        if version.has_name and not noname:
            if name == "":
                name = gen_name(e.name, False)
            o += ",name=" + name + "_NUM"
        if newextra:
            o += "," + ",".join(newextra)
        o += "/"

        # explode boxes if needed
        def box_name(n):
            return "%s_%d" % (e.unit, n)
        def box_n_exists(n):
            if one_unit and n > 0:
                return False
            return box_exists(box_name(n))
        if not noexplode and not box_exists(e.unit) and box_n_exists(0):
            return ",".join(["uncore_" + box_name(x) + o.replace("_NUM", "_%d" % (x)) for x in
                             itertools.takewhile(box_n_exists, itertools.count())])
        return "uncore_%s%s" % (e.unit, o.replace("_NUM", ""))

    def filter_qual(self):
        def check_qual(q):
            if q == "":
                return False
            if q == "one_unit":
                return True
            if "=" in q:
                q, _ = q.split("=")
            if has_format_any(q, "uncore_" + self.unit):
                return True
            warn_once("%s: format %s not supported. Filtering out" % (self.unit, q))
            return False

        self.newextra = ",".join(filter(check_qual, convert_uncore(self.newextra, ()).split(",")))
    output = output_newstyle

def ffs(flag):
    assert flag != 0
    m = 1
    j = 0
    while (flag & m) == 0:
        m = m << 1
        j += 1
    return j

perf_qual = "kuhGHSD"  # without pebs

def extra_set(e):
    return set(map(lambda x: x[0],
        re.findall(r"(" + "|".join([x[0] for x in qual_map + qualval_map + uncore_map]) + "|[" + perf_qual + "]|p+)", e)))

def merge_extra(a, b):
    m = a | b
    if 'ppp' in m:
        m = m - set(['p', 'pp'])
    if 'pp' in m:
        m = m - set(['p'])
    m = m - set([':'])
    return m

def print_event(name, desc, f, human, wrap, pmu=""):
    desc = "".join([y for y in desc if y < chr(127)])
    print("  %-42s" % (name,), end='', file=f)
    if pmu:
        print(" [%s]" % pmu, end='', file=f)
    if human:
        print("\n%s" % (wrap.fill(desc),), file=f)
    else:
        print(" [%s]" % (desc,), file=f)

def uncore_exists(box, postfix=""):
    if file_exists("/sys/bus/event_source/devices/uncore_" + box + postfix):
        return True
    if file_exists("/sys/bus/event_source/devices/uncore_" + box + "_0" + postfix):
        return True
    return False

missing_boxes = set()

def check_uncore_event(e, extramsg):
    if uncore_exists(e.unit):
        if e.cmask and not uncore_exists(e.unit, "/format/cmask"):
            warn_once("Uncore unit " + e.unit + " missing cmask for " + e.name)
            extramsg.append("not supported due to missing cmask in PMU")
            return None
        if e.umask and not uncore_exists(e.unit, "/format/umask"):
            warn_once("Uncore unit " + e.unit + " missing umask for " + e.name)
            extramsg.append("not supported due to missing umask in PMU")
            return None
        return e
    if e.unit not in missing_boxes:
        warn_once("Uncore unit " + e.unit + " missing")
        missing_boxes.add(e.unit)
    extramsg.append("not supported due to missing PMU")
    return None

fixed_counters = {
    "inst_retired.any": (0xc0, 0, 0),
    "cpu_clk_unhalted.thread": (0x3c, 0, 0),
    "cpu_clk_unhalted.thread_any": (0x3c, 0, 1),
    "cpu_clk_unhalted.core": (0x3c, 0, 1),
}

def update_ename(ev, name):
    if ev:
        ev = copy.deepcopy(ev)
        ev.name = name
    return ev

def json_open(name):
    if ocverbose:
        print("open", name, file=sys.stderr)
    d = open(name, "rb").read()
    if not isinstance(d, str):
        d = d.decode('utf-8')

    json_data = json.loads(d)
    if isinstance(json_data, dict) and 'Events' in json_data:
        json_data = json_data['Events']

    return json_data

class EmapNativeJSON(object):
    """Read an event table."""

    def __init__(self, name, pmu):
        self.events = {}
        self.perf_events = {}
        self.codes = {}
        self.desc = {}
        self.pevents = {}
        self.latego = False
        self.uncore_events = {}
        self.error = False
        self.pmu = pmu
        self.name = name
        self.read_events(name)

    def add_event(self, e):
        self.events[e.name] = e
        self.perf_events[e.name.replace('.', '_')] = e  # workaround for perf-style naming
        if e.pname:
            self.pevents[e.pname] = e
        self.codes[e.val] = e
        self.desc[e.name] = e.desc
        e.pmu = self.pmu

    def read_table(self, r):
        for row in r:
            def get(x):
                if isinstance(row[x], bool):
                    return "1" if row[x] else "0"
                return row[x]
            def gethex(x):
                return int(get(x).split(",")[0], 16)
            def getdec(x):
                return int(get(x), 10)

            name = get(u'EventName').lower().rstrip()
            try:
                code = gethex(u'EventCode')
                umask = gethex(u'UMask')
                if 'UMaskExt' in row:
                    umask |= gethex(u'UMaskExt') << (40 - 8)
            except ValueError:
                if ocverbose:
                    print("cannot parse event", name)
                continue
            anyf = 0
            if name in fixed_counters:
                code, umask, anyf = fixed_counters[name]
            if u'Other' in row:
                other = gethex(u'Other') << 16
            else:
                other = 0
            other |= gethex(u'EdgeDetect') << 18
            if u'AnyThread' in row:
                other |= (gethex(u'AnyThread') | anyf) << 21
            if u'CounterMask' in row:
                other |= getdec(u'CounterMask') << 24
            elif u'CMask' in row:
                other |= getdec(u'CMask') << 24
            other |= gethex(u'Invert') << 23
            val = code | (umask << 8) | other
            val &= EVMASK
            d = get(u'PublicDescription')
            if d is None:
                d = ''
            d = d.strip()
            e = Event(name, val, d)
            e.newextra = ""
            if other & ((1 << 16)|(1 << 17)):
                if other & (1<<16):
                    e.extra += "u"
                if other & (1 << 17):
                    e.extra += "k"
            e.perfqual = None
            if u'MSRIndex' in row and get(u'MSRIndex') and get(u'MSRValue'):
                msrnum = gethex(u'MSRIndex')
                msrval = gethex(u'MSRValue')
                if version.offcore and msrnum in (0x1a6, 0x1a7):
                    e.newextra = ",offcore_rsp=0x%x" % (msrval, )
                    e.perfqual = "offcore_rsp"
                elif version.ldlat and msrnum in (0x3f6,):
                    e.newextra = ",ldlat=0x%x" % (msrval, )
                    e.perfqual = "ldlat"
                elif msrnum == 0x3f7:
                    e.newextra = ",frontend=%#x" % (msrval, )
                    e.perfqual = "frontend"
                # add new msr here
                else:
                    e.msrval = msrval
                    e.msr = msrnum
            if u'SampleAfterValue' in row:
                e.overflow = get(u'SampleAfterValue')
            e.counter = get(u'Counter')
            e.precise = getdec(u'Precise') if u'Precise' in row else 0
            e.collectpebs = getdec(u'CollectPEBS') if u'CollectPEBS' in row else 0
            e.pebs = getdec(u'PEBS') if u'PEBS' in row else 0
            if e.collectpebs > 1 or e.pebs >= 2:
                e.extra += "pp"
                if len(e.counter.split(",")) == 1:
                    e.extra += "p"
            try:
                if get(u'Errata') != "null":
                    try:
                        d += " Errata: "
                        d += get(u'Errata')
                        e.errata = get(u'Errata')
                    except UnicodeDecodeError:
                        pass
            except KeyError:
                pass
            e.desc = d
            for (flag, name) in extra_flags:
                if val & flag:
                    e.newextra += ",%s=%d" % (name, (val & flag) >> ffs(flag), )
            e.period = int(get(u'SampleAfterValue')) if u'SampleAfterValue' in row else 0
            self.add_event(e)

    def getevent(self, e, nocheck=False, extramsg=[]):
        """Retrieve an event with name e. Return Event object or None.
           When nocheck is set don't check against current system."""
        e = e.lower()
        extra = ""
        edelim = ""
        m = re.match(r'([^:]+):request=([^:]+):response=([^:]+)', e)
        if m:
            ename = m.group(1) + "." + m.group(2) + "." + m.group(3)
            return update_ename(self.getevent(ename, nocheck=nocheck, extramsg=extramsg), e)
        m = re.match(r'(.*?):(.*)', e)
        if m:
            extra = m.group(2)
            edelim = ":"
            e = m.group(1)
        if e in self.events:
            # hack for now. Avoid ambiguity with :p
            # Should handle qualmap properly here
            extra = extra.replace("period=", "sample-after=")
            userextra = extra
            extra = extra_set(extra)
            ev = self.events[e]
            ev_extra = extra_set(ev.extra)
            if extra and merge_extra(ev_extra, extra) > ev_extra:
                ev = copy.deepcopy(self.events[e])
                ev.userextra = userextra
                ev.extra = "".join(sorted(merge_extra(ev_extra, extra)))
                return ev
            return self.events[e]
        elif e.endswith("_ps"):
            return update_ename(self.getevent(e[:-3] + ":p" + extra), e)
        elif e.startswith("offcore") and (e + "_0") in self.events:
            return update_ename(self.getevent(e + "_0" + edelim + extra), e)
        elif e in self.uncore_events:
            ev = self.uncore_events[e]
            if ev and not nocheck:
                ev = check_uncore_event(ev, extramsg)
            if ev and extra:
                ev = copy.deepcopy(ev)
                ev.newextra = extra
            return ev
        elif e in self.perf_events:
            return self.perf_events[e]
        elif e in self.pevents:
            return self.pevents[e]

        extramsg.append("event not found for %s for %s" % (self.pmu, self.name))
        return None

    # XXX need to handle exploded events
    def update_event(self, e, ev):
        if e not in self.pevents:
            self.pevents[e] = ev

    def getraw(self, r):
        e = "r%x" % (r)
        if e in self.pevents:
            ev = self.pevents[e]
            s = ev.name
            if ev.extra:
                s += ":" + ev.extra
            return s
        return "!Raw 0x%x" % (r,)

    def getperf(self, p):
        if p in self.pevents:
            e = self.pevents[p]
            n = e.name
            if e.userextra:
                n += ":" + e.userextra
            return n
        return p

    def dumpevents(self, f=sys.stdout, human=True, uncore=True):
        """Print all events with descriptions to the file descriptor f.
           When human is true word wrap all the descriptions."""
        wrap = None
        if human:
            wrap = textwrap.TextWrapper(initial_indent="     ",
                                        subsequent_indent="     ")
        for k in sorted(self.events.keys()):
            print_event(k, self.desc[k], f, human, wrap, self.pmu)
        if uncore:
            for k in sorted(self.uncore_events.keys()):
                print_event(k, self.uncore_events[k].desc, f, human, wrap)

    def read_events(self, name):
        """Read JSON normal events table."""
        if name.find("JKT") >= 0 or name.find("Jaketown") >= 0:
            self.latego = True
        try:
            data = json_open(name)
        except ValueError as e:
            print("Cannot parse", name + ":", e.message, file=sys.stderr)
            self.error = True
            return
        self.read_table(data)
        if "topdown.slots" in self.events:
            self.add_topdown()

    def add_offcore(self, name):
        """Read offcore table."""
        data = json_open(name)
        #   {
        #    "MATRIX_REQUEST": "DEMAND_DATA_RD",
        #    "MATRIX_RESPONSE": "NULL",
        #    "MATRIX_VALUE": "0x0000000001",
        #    "MATRIX_REGISTER": "0,1",
        #    "DESCRIPTION": "Counts demand data reads that"
        #   },

        offcore_response = self.getevent("OFFCORE_RESPONSE")
        if not offcore_response:
            return
        requests = []
        responses = []

        for row in data:
            if row[u"MATRIX_REQUEST"].upper() != "NULL":
                requests.append((row[u"MATRIX_REQUEST"], row[u"MATRIX_VALUE"], row[u"DESCRIPTION"]))
            if row[u"MATRIX_RESPONSE"].upper() != "NULL":
                responses.append((row[u"MATRIX_RESPONSE"], row[u"MATRIX_VALUE"], row[u"DESCRIPTION"]))

        def create_event(req_name, req_val, req_desc, res_name, res_val, res_desc):
            oe = copy.deepcopy(offcore_response)
            oe.name = ("OFFCORE_RESPONSE.%s.%s" % (req_name, res_name)).lower()
            if oe.name.lower() in self.events:
                return
            oe.msrval = int(req_val, 16) | (int(res_val, 16) << 16)
            oe.desc = req_desc + " " + res_desc
            if version.offcore:
                oe.newextra = ",offcore_rsp=0x%x" % (oe.msrval, )
            else:
                oe.msr = 0x1a6
            self.add_event(oe)

        for a, b in itertools.product(requests, responses):
            create_event(*(a + b))

    def add_uncore(self, name, force=False):
        data = json_open(name)
        for row in data:
            name = row['EventName'].lower()
            try:
                self.uncore_events[name] = UncoreEvent(name, row)
            except UnicodeEncodeError:
                pass

    def add_topdown(self):
        def td_event(name, pname, desc, counter):
            e = Event(name, 0, desc)
            e.counter = counter
            e.pname = pname
            self.add_event(e)
        td_event("perf_metrics.retiring", "topdown-retiring", "Number of slots the pipeline was frontend bound.", "32")
        td_event("perf_metrics.bad_speculation", "topdown-bad-spec", "Number of slots the pipeline was doing bad speculation.", "33")
        td_event("perf_metrics.frontend_bound", "topdown-fe-bound", "Number of slots the pipeline was frontend bound.", "34")
        td_event("perf_metrics.backend_bound", "topdown-be-bound", "Number of slots the pipeline was backend bound.", "35")
        td_event("topdown.slots", "slots", "Number of slots", "36")

        if "topdown.memory_bound_slots" in self.events:
            td_event("perf_metrics.heavy_operations", "topdown-heavy-ops", "Number of slots pipeline retired microcode instructions with >2 uops", "36")
            td_event("perf_metrics.branch_mispredicts", "topdown-br-mispredict", "Number of slots frontend was bound by branch mispredictions", "37")
            td_event("perf_metrics.memory_bound", "topdown-mem-bound", "Number of slots backend was bound by memory", "38")
            td_event("perf_metrics.fetch_latency", "topdown-fetch-lat", "Number of slots frontend was bound by memory fetch latency", "39")

pmu_to_type = {
    "cpu_core": ("hybridcore", "Core"),
    "cpu_atom": ("hybridcore", "Atom"),
    "cpu": ("core", None),
}

def json_with_extra(el, eventmap_is_file, pmu, typ_key):
    typ = pmu_to_type[pmu]
    if typ_key:
        typ = (typ[0], typ_key)
    name = event_download.eventlist_name(el, key=typ[0], hybridkey=typ[1])
    if not os.path.exists(name):
        name = event_download.eventlist_name(el, "hybridcore", hybridkey=typ[1])
    emap = EmapNativeJSON(name, pmu)
    if not emap or emap.error:
        print("parsing", name, "failed", file=sys.stderr)
        return None
    if experimental:
        try:
            emap.read_events(event_download.eventlist_name(el, "core experimental"))
        except IOError:
            pass
    add_extra_env(emap, el, eventmap_is_file)
    return emap

def add_extra_env(emap, el, eventmap_is_file):
    try:
        oc = os.getenv("OFFCORE")
        if oc:
            oc = canon_emapvar(oc, "matrix")
            oc = event_download.eventlist_name(el, "offcore")
            emap.add_offcore(oc)
        elif not eventmap_is_file:
            oc = event_download.eventlist_name(el, "offcore")
            if os.path.exists(oc) and el != oc:
                emap.add_offcore(oc)
            if experimental:
                oc = event_download.eventlist_name(el, "offcore experimental")
                if os.path.exists(oc) and oc != el:
                    emap.add_offcore(oc)
    except IOError:
        print("Cannot open offcore", oc, file=sys.stderr)
    try:
        uc = os.getenv("UNCORE")
        if uc:
            uc = canon_emapvar(uc, "uncore")
            uc = event_download.eventlist_name(uc, "uncore")
            emap.add_uncore(uc)
        elif not eventmap_is_file:
            uc = event_download.eventlist_name(el, "uncore")
            if os.path.exists(uc) and uc != el:
                emap.add_uncore(uc)
            if experimental:
                uc = event_download.eventlist_name(el, "uncore experimental")
                if os.path.exists(uc) and uc != el:
                    emap.add_uncore(uc)
    except IOError:
        print("Cannot open uncore", uc, file=sys.stderr)
    def read_map(env, typ, r):
        try:
            e2 = os.getenv(env)
            if e2:
                e2 = canon_emapvar(e2, typ)
                r(e2)
                # don't try to download for now
        except IOError:
            print("Cannot open " + env, e2, file=sys.stderr)
    read_map("EVENTMAP2", "core", emap.read_events)
    read_map("EVENTMAP3", "core", emap.read_events)
    read_map("UNCORE2", "uncore", emap.add_uncore)

def canon_emapvar(el, typ):
    if ("*" in el or "." in el or "_" in el) and "/" not in el and not file_exists(el):
        el = "%s/%s" % (event_download.getdir(), el)
    if '*' in el:
        l = glob.glob(el)
        if l:
            if len(l) > 1:
                l = [x for x in l if x.find(typ) >= 0]
            if l:
                el = l[0]
    return el

def find_emap(eventvar="EVENTMAP", pmu="cpu", typ=None):
    """Search and read a perfmon event map.
       When the EVENTMAP environment variable is set read that, otherwise
       read the map for the current CPU. EVENTMAP can be a CPU specifier
       in the map file or a path name.
       Dito for the OFFCORE and UNCORE environment variables.

       Optionally pass the name of the EVENTMAP variable, and the cpu pmu name,
       and the cpu hybrid type in the map file.

       Return an emap object that contains the events and can be queried
       or None if nothing is found or the current CPU is unknown."""
    el = os.getenv(eventvar)
    if not el:
        eventmap_is_file = False
        el = event_download.get_cpustr()
    else:
        eventmap_is_file = "/" in el
    el = canon_emapvar(el, "core")
    if "/" in el:
        try:
            emap = EmapNativeJSON(el, pmu)
            if not emap or emap.error:
                return None
            add_extra_env(emap, el, eventmap_is_file)
            return emap
        except IOError:
            return None
    try:
        if not force_download:
            return json_with_extra(el, eventmap_is_file, pmu, typ)
    except IOError:
        pass
    try:
        toget = ["core"]
        if not os.getenv("OFFCORE"):
            toget.append("offcore")
        if not os.getenv("UNCORE"):
            toget.append("uncore")
        if experimental:
            toget += [x + " experimental" for x in toget]
        event_download.download(el, toget)
        return json_with_extra(el, eventmap_is_file, pmu, typ)
    except IOError:
        pass
    return None

def process_events(event, print_only, period, noexplode):
    overflow = None
    # replace inner commas so we can split events
    event = re.sub(r"([a-z][a-z0-9]+/)([^/]+)/",
            lambda m: m.group(1) + m.group(2).replace(",", "#") + "/",
            event)
    el = event.split(",")
    nl = []
    group_index = 0
    for i in el:
        group_start = ""
        group_end = ""
        start = ""
        end = ""
        if i.startswith('{'):
            group_start = "{"
            i = i[1:]
            group_index = len(nl)
        m = re.match(r'(.*)(\}(:.*)?)', i)
        if m:
            group_end = m.group(2)
            i = m.group(1)
        i = i.strip()
        m = re.match(r'([^/]+)/([^#]+)(#?.*?)/(.*)', i)
        if m:
            start = m.group(1) + "/"
            for emap in emap_list:
                if emap.pmu == m.group(1):
                    ev = emap.getevent(m.group(2))
                    break
            else:
                ev = emap_list[0].getevent(m.group(2))
            end = m.group(3) + "/"
            if ev:
                qual = "".join(sorted(merge_extra(extra_set(ev.extra), extra_set(m.group(4)))))
                end += qual
                i = ev.output_newstyle(period=period, noexplode=noexplode)
                if i.count("/") > 2: # was it exploded?
                    start = ""
                    end = ""
            else:
                start = ""
                end = ""
        else:
            ev = None
            res = [x.getevent(i) for x in emap_list]
            res = [x for x in res if x]
            if res:
                if len(res) > 1 and not i.lower().startswith("unc_"):
                    print("Event %s is not unique on hybrid CPUs. Add cpu_*// prefixes" % i, file=sys.stderr)
                ev = res[0]
                i = ev.output(period=period, noexplode=noexplode)
        if ev:
            if ev.msr:
                msr.checked_writemsr(ev.msr, ev.msrval, print_only)
            for emap in emap_list:
                if emap.latego and (ev.val & 0xffff) in latego.latego_events:
                    latego.setup_event(ev.val & 0xffff, 1)
            overflow = ev.overflow
        event = (group_start + start + i + end + group_end).replace("#", ",")
        nl.append(event)
        if ev:
            emap.update_event(event, ev)
        if "S" in group_end:
            for j in range(group_index + 1, len(nl)):
                nl[j] = re.sub(r',period=\d+', '', nl[j])

    return str.join(',', nl), overflow

def getarg(i, cmd):
    if sys.argv[i][2:] == '' or sys.argv[i][:2] == '--':
        cmd.append(sys.argv[i])
        i += 1
        arg = ""
        if len(sys.argv) > i:
            arg = sys.argv[i]
        prefix = ""
    else:
        arg = sys.argv[i][2:]
        prefix = sys.argv[i][:2]
    return arg, i, prefix

def process_args():
    perf = os.getenv("PERF")
    if not perf:
        perf = "perf"
    cmd = [perf]

    noexplode = False
    overflow = None
    print_only = False
    never, no, yes = range(3)
    record = no
    i = 1
    while i < len(sys.argv):
        if sys.argv[i] == "--print":
            print_only = True
        elif sys.argv[i] == "--force-download":
            pass
        elif sys.argv[i] == "--experimental":
            pass
        elif sys.argv[i] == "--no-period":
            record = never
        elif sys.argv[i] == "--noexplode":
            noexplode = True
        elif sys.argv[i] == "record" and record == no:
            cmd.append(sys.argv[i])
            record = yes
        elif sys.argv[i][0:2] == '-e' or sys.argv[i] == '--event':
            event, i, prefix = getarg(i, cmd)
            event, overflow = process_events(event, print_only,
                                             record == yes,
                                             noexplode)
            cmd.append(prefix + event)
        elif record and (sys.argv[i][0:2] == '-c' or sys.argv[i] == '--count'):
            oarg, i, prefix = getarg(i, cmd)
            if oarg == "default":
                if overflow is None:
                    print("""
Specify the -e events before -c default or event has no overflow field.""", file=sys.stderr)
                    sys.exit(1)
                cmd.append(prefix + overflow)
            else:
                cmd.append(prefix + oarg)
        else:
            cmd.append(sys.argv[i])
        i += 1
    print(" ".join(map(quote, cmd)))
    if print_only:
        sys.exit(0)
    return cmd

if sys.version_info.major == 3:
    popentext = dict(text=True)
else:
    popentext = {}

def get_pager():
    f = sys.stdout
    if f.isatty():
        try:
            sp = subprocess.Popen(["less", "-F"], stdin=subprocess.PIPE, **popentext) # type: ignore
            return sp.stdin, sp
        except OSError:
            f = sys.stdout
    return f, None

def perf_cmd(cmd):
    if len(emap_list) == 0:
        sys.exit(subprocess.call(cmd))
    elif len(sys.argv) >= 2 and sys.argv[1] == "list":
        pager, proc = get_pager()
        try:
            l = subprocess.Popen(cmd, stdout=pager)
            l.wait()
            print(file=pager)
            uncore = True
            for emap in emap_list:
                if len(emap_list) > 1:
                    print("%s:\n" % emap.pmu, file=pager)
                emap.dumpevents(pager, proc is not None, uncore)
                uncore = False
            if proc:
                pager.close()
                proc.wait()
        except IOError:
            pass
    elif len(sys.argv) >= 2 and (sys.argv[1] == "report" or sys.argv[1] == "stat"):
        direct = version.has_name
        if not direct:
            for w in sys.argv:
                if w == "--tui":
                    direct = True
                    break
        if direct:
            ret = subprocess.call(cmd)
            latego.cleanup()
            sys.exit(ret)
        try:
            pipe = subprocess.Popen(cmd,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT).stdout
            raw = lambda e: " " + emap.getraw(int(e.group(1), 16))
            for i in pipe:
                i = re.sub("[rR]aw 0x([0-9a-f]{4,})", raw, i)
                i = re.sub("r([0-9a-f]{4,})", raw, i)
                i = re.sub("(cpu(_core|_atom)?/.*?/)", lambda e: emap.getperf(e.group(1)), i)
                print(i, end='')
        except IOError:
            pass
        pipe.close()
        latego.cleanup()
    else:
        sys.exit(subprocess.call(cmd))

def find_pmus():
    g = glob.glob("/sys/bus/event_source/devices/cpu*")
    if len(g) == 0:
        g = ["/sys/bus/event_source/devices/cpu"]
    return [i.replace("/sys/bus/event_source/devices/", "") for i in g]


def main():
    global force_download, experimental, noexplode
    global msr
    for j in sys.argv:
        if j == "--force-download":
            force_download = True
        if j == "--experimental":
            experimental = True
        if j == "--noexplode":
            noexplode = True
    msr = MSR()
    pmus = find_pmus()
    for pmu in pmus:
        emap = find_emap(pmu=pmu)
        if not emap:
            print("Do not recognize CPU or cannot find CPU map file.", file=sys.stderr)
        else:
            emap_list.append(emap)
    cmd = process_args()
    try:
        perf_cmd(cmd)
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()




# Copyright (c) 2013, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# do some (not so) simple expression parsing, mainly to handle the
# with:... construct look up the sub events and convert them into a
# function call we also expand sub expressions (references to other
# events), but we need to recurse on () and for sub exprs
# this does not require handling full operator precedence
#
from __future__ import print_function
import re
import inspect

import ucevent
import ucmsg

dbg = ucmsg.debug_msg

class ParseError(Exception):
    def __init__(self, msg):
        self.msg = msg

def is_id(t):
    return re.match(r'[a-zA-Z][\[\]:a-zA-Z0.9.]*', t)

def is_number(t):
    return re.match(r"(0x)?[0-9]+", t)

cpu_events = {
    "INST_RETIRED.ALL": "instructions",
    "TOTAL_CORE_CYCLES": "cycles",
}

def event_expr(v, n):
    return "EV('%s', '%s')" % (v, n)

def fix_token(t, box, user_mode):
    if ((t.find(".") >= 0 and re.match(r"[A-Z][0-9a-zA-Z_.]{5,}", t))
            or re.match(r"[A-Z]", t)
            or t.startswith("iMC")
            or t.startswith("R2PCIe")):
        if t in cpu_events:
            return event_expr(cpu_events[t], t)
        if t in ucevent.cpu_aux.alias_events:
            return event_expr(ucevent.cpu_aux.alias_events[t], t)
        if user_mode:
            m = re.match(r"([^.]+)\.", t)
            if m:
                box = m.group(1)
        name = t
        if not t.startswith(box + "."):
            name = box + "." + t
        ev = ucevent.lookup_event(name)
        if not ev:
            ev = ucevent.lookup_event(t)
        if ev:
            if "Equation" in ev:
                tl = tokenize(ev["Equation"], box)
                n = expr(tl)
                return n
            evl = ucevent.format_event(ev)
            n = event_expr(evl[0], name)
            if len(evl) > 1:
                n = n.replace("_0/", "_INDEX/")
            return n
    return t

# really poor man's tokenizer
def tokenize(eq, box, user_mode=False):
    eq = eq.replace(".{", " .{")
    eq = eq.replace("with:", "with: ")
    eq = eq.replace("(on Core)", "")
    eq = re.sub(r"([*={}(),/<>]|>=|==|<=)", r" \1 ", eq)
    eq = re.sub(r"\.([a-z])", r" . \1", eq)
    eq = re.sub(r"([=<>])  \1", r"\1\1", eq)
    eq = re.sub(r"([<>])  =", r"\1=", eq)
    return list(map(lambda x: fix_token(x, box, user_mode), eq.split()))

# expand event lists to multiple boxes after parsing
# this avoids having to pass this all around the parser
# we assume all are for the same box
def expand_events(s):
    m = re.search(r"uncore_([^/]+)_INDEX", s)
    if not m:
        return [s]
    l = []
    for j in ucevent.find_boxes(m.group(1)):
        m = re.search(r'_(\d+)$', j)
        if m:
            l.append(s.replace("_INDEX", "_" + m.group(1)))
    if len(l) == 0:
        return [s]
    return l

def expect(tl, c):
    dbg("expect", "expecting %s" % (c))
    if not tl:
        raise ParseError("expected %s, got end" % (c))
    if tl[0] != c:
        raise ParseError("expected %s instead of %s: %s" % (c, tl[0], tl[:5]))
    return tl[1:]

def convert_qual(q, v):
    qual_alias = ucevent.cpu_aux.qual_alias
    if q in qual_alias:
        return (qual_alias[q], v)
    m = re.match(r"([A-Za-z0-9_]+)\[(\d+):(\d+)\]", q)
    if m:
        q = m.group(1)
        val = (int(v, 0) << int(m.group(3)))
        v = "%#x" % (val)
    if q in qual_alias:
        return (qual_alias[q], v)
    return (q, v)

def is_ev(l):
    return isinstance(l, str) and l.startswith("EV(")

def apply_expr(o, fl, vl):
    if is_list(o):
        return list(map(lambda x: apply_expr(x, fl, vl), o))
    if is_ev(o):
        for f, v in zip(fl, vl):
            nn, nv = convert_qual(f, v)
            # existing qualifier to or it to?
            m = re.search(r",%s=([^,/]+)" % (nn), o)
            if m:
                val = int(m.group(1), 0)
                val |= int(nv, 0)
                o = re.sub(r",(%s)=([^,/]+)", r"\1=%#x," % (val), o)
            else:
                add = ",%s=%s" % (nn, nv)
                o = o.replace("/'", add + "/'")
    return o

def has_ev(l):
    if is_list(l):
        for j in l:
            if has_ev(j):
                return True
    elif is_ev(l):
        return True
    return False

def apply_list(o, fl, vl):
    dbg("expr", "apply %s and %s to %s from %s" %
                   (fl, vl, o, inspect.stack()[1][2:]))
    for j in fl:
        if is_list(j):
            for k in j:
                n = k.split('=')
                o = apply_expr(o, [n[0]], [n[1]])
    fl = list(filter(lambda x: not is_list(x), fl))
    if len(fl) != len(vl):
        print("MISMATCHED APPLY",fl,vl,o,inspect.stack()[1][2:])
        return o
    if not has_ev(o):
        evo = []
        for f, v in zip(fl, vl):
            if is_id(f):
                f = "." + f
            evo.append("%s%s=%s" % (o, f, v))
        return evo
    return apply_expr(o, fl, vl)

closener = { "{": "}", "(": ")" }

# a = NUMBER
# a
# NUMBER
# [...]           (???)
# a { list }
def parse_term(tl):
    name = tl[0]
    if not (is_id(name) or is_number(name) or name[0] == '['):
        raise ParseError("expected identifier or number in list not %s: %s" %
                            (tl[0], tl[:5]))
    tl = tl[1:]
    if tl[0] == '.':
        tl = tl[1:]
        name = name + "." + tl[0]
        tl = tl[1:]
    if tl[0] == '=':
        tl = tl[1:]
        if not is_number(tl[0]):
            raise ParseError("expected number after =, not %s" % (tl[0]))
        return name, tl[0], tl[1:]
    if tl[0] == '{':
        fl, vl, tl = parse_list(tl[1:])
        tl = expect(tl, '}')
        if tl[0] == '=':
            tl = tl[1:]
            if tl[0] not in ('(', '{'):
                raise ParseError("expect ( or {, not %s" % (tl[:5]))
            opener = tl[0]
            vl, _, tl = parse_list(tl[1:])
            tl = expect(tl, closener[opener])
        name = apply_list(name, fl, vl)
    return name, None, tl

# term { , term }
def parse_list(tl):
    ls = []
    nm = []
    while True:
        term, nmt, tl = parse_term(tl)
        ls.append(term)
        if nmt is not None:
            nm.append(nmt)
        if tl[0] != ',':
            break
        tl = tl[1:]
    return ls, nm, tl


# '{' list '}'
# '{' list '}' = ( val_list)
# a = NUMBER
# a . '{' list '}' = ( val_list )
# a . x = NUMBER
def parse_with(tl, orig):
    #print("with",tl)
    if tl[0] == '{':
        ls, nm, tl = parse_list(tl[1:])
        vl = []
        tl = expect(tl, '}')
        if tl and tl[0] == '=':
            tl = tl[1:]
            if tl[0] in ('(', '{'):
                opener = tl[0]
                nm, _, tl = parse_list(tl[1:])
                tl = expect(tl, closener[opener])
            else:
                raise ParseError("expected { or (, not %s" % (tl[:5]))
        orig = apply_list(orig, ls, nm)
        return orig, tl
    # id
    if is_id(tl[0]) or tl[0][0] == '[':
        idn = tl[0]
        tl = tl[1:]
        if tl[0] == '.':
            tl = tl[1:]
            # id . id
            if is_id(tl[0]):
                id2 = tl[0]
                tl = tl[1:]
                # id . id = NUMBER
                if tl[0] == '=':
                    tl = tl[1:]
                    if not is_number(tl[0]):
                        raise ParseError("expecting number after =, not %s" % (tl[:5]))
                    number = tl[0]
                    tl = tl[1:]
                else:
                    number = "1"
                orig = apply_list(orig, [idn + "." + id2], [number])
                return orig, tl
            # .{ list } = { list }
            tl = expect(tl, '{')
            ls, vl, tl = parse_list(tl)
            tl = expect(tl, '}')
            if tl[0] == '=':
                tl = tl[1:]
                if tl[0] == '(' or tl[0] == '{':
                    opener = tl[0]
                    vl, _, tl = parse_list(tl[1:])
                    tl = expect(tl, closener[opener])
            orig = apply_list(orig, ls, vl)
            return orig, tl
        tl = expect(tl, '=')
        num = tl[0]
        if not is_number(num):
            raise ParseError("expected number, not %s" % (num))
        orig = apply_list(orig, [idn], [num])
        return orig, tl[1:]
    elif tl[0] == '.':
        tl = tl[1:]
        if is_id(tl[0]):
            # id . a = NUMBER
            vl = [tl[0]]
            tl = tl[1:]
            if tl[0] == '=':
                tl = tl[1:]
                num = tl[0]
                if not is_number(num):
                    raise ParseError("Expected number, not %s" % (tl[:5]))
            else:
                num = 1
            vl = [num]
        else:
            tl = expect(tl, '{')
            fl, _, tl = parse_list(tl)
            tl = expect(tl, '}')
            tl = expect(tl, '=')
            tl = expect(tl, '(')
            vl, _, tl = parse_list(tl)
            tl = expect(tl, ')')
        orig = apply_list(orig, fl, vl)
        return orig, tl
    raise ParseError("parse error in with clause at %s (%s)" % (tl[0], tl))
    return [], []

operators = ("+", "-", "*", "/", ">", "<", ">=", "<=", "==", "<<", ">>")

def expr_term(tl):
    if tl[0] == '(':
        out, tl = expr(tl[1:])
        if isinstance(out, list):
            out = [out]
        tl = expect(tl, ')')
    else:
        out = [tl[0]]
        tl = tl[1:]
        if tl and tl[0] == '{':
            fl, vl, tl = parse_list(tl[1:])
            tl = expect(tl, '}')
            out = apply_list(out, fl, vl)
    if not tl:
        return out, tl
    if tl and tl[0] == 'with:':
        out, tl = parse_with(tl[1:], out)
    return out, tl

def expr(tl):
    out = []
    while True:
        no, tl = expr_term(tl)
        out += no
        if tl and tl[0] in operators:
            out.append(tl[0])
            tl = tl[1:]
        else:
            break
    return out, tl

def is_list(e):
    return isinstance(e, list) or isinstance(e, tuple)

def expr_flat(e):
    if not is_list(e):
        return e
    l = []
    for i in e:
        s = expr_flat(i)
        if is_list(s):
            if len(s) > 0:
                l.append('(')
                l += s
                l.append(')')
        else:
            l.append(s)
    return l

def apply_one_user_qual(x, qual):
    if not x.startswith("EV("):
        return x
    if any(map(lambda r: re.search(r, x), ucevent.cpu_aux.clockticks)):
        return x
    return x.replace("/'", "," + qual + "/'")

def apply_user_qual(e, qual):
    return list(map(lambda x: apply_one_user_qual(x, qual), e))

def parse(s, box, quiet=False, user_mode=False, qual=None):
    try:
        if not quiet:
            print("Expression", s)
        tl = tokenize(s, box, user_mode)
        dbg("tokenize", tl)
        e, tl = expr(tl)
        if len(tl) > 0:
            raise ParseError("unexpected token %s at end" % (tl[0]))
        dbg("expr", e)
        eflat = expr_flat(e)
        if qual:
            eflat = apply_user_qual(eflat, qual)
        res = " ".join(eflat)
        res = expand_events(res)
        dbg("expanded", tl)
        return res
    except ParseError as p:
        print("PARSE-ERROR", p.msg, s)
        return []

if __name__ == '__main__':
    assert is_id("x")
    print(parse("a + b + c", "foo"))
    print(parse("a with:x=1 + (b with:.{o}=(1) + c with:{edge}) with:{x=1,y=2} + d", "foo"))




#!/usr/bin/env python
# A description of the perf.data file format in "construct"
#
# Copyright (c) 2011-2013, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# Only works on Little-Endian with LE input files. Sorry.
#
# TBD:
# Generic Bitfield adapter that handles endian properly?
# check size in all cases (or use optional+tunnel)
# tracing support
# processor trace (aux) data and extra info headers
# sample_id is not handled correctly in all cases
#
from __future__ import print_function
from construct import (If, Embedded, Struct, SNInt32, UNInt64, Flag, BitStruct,
                       Padding, Enum, Array, Bytes, Anchor, UNInt32, GreedyRange,
                       CString, PrefixedArray, TunnelAdapter, Magic, Pointer,
                       BitField, UNInt16, HexDumpAdapter, String, Pass, Value,
                       Switch)

def sample_type(ctx):
    return ctx.attr.perf_event_attr.sample_type

def sample_id_size(ctx):
    st = sample_type(ctx)
    return (st.tid + st.time + st.id + st.stream_id + st.cpu +
            st.identifier) * 8

def sample_id():
    return If(lambda ctx: True,  # xxx check size
              Embedded(Struct("id_all",
                              If(lambda ctx: sample_type(ctx).tid,
                                 Embedded(Struct("pid2",
                                                 SNInt32("pid2"),
                                                 SNInt32("tid2")))),
                              If(lambda ctx: sample_type(ctx).time,
                                 UNInt64("time2")),
                              If(lambda ctx: sample_type(ctx).id,
                                 UNInt64("id2")),
                              If(lambda ctx: sample_type(ctx).stream_id,
                                 UNInt64("stream_id2")),
                              If(lambda ctx: sample_type(ctx).cpu,
                                 Embedded(Struct("cpu",
                                                 UNInt32("cpu2"),
                                                 UNInt32("res")))),
                              If(lambda ctx: sample_type(ctx).identifier,
                                 UNInt64("identifier2")))))

def fork_exit(name):
    return Struct(name,
                  SNInt32("pid"),
                  SNInt32("ppid"),
                  SNInt32("tid"),
                  SNInt32("ptid"),
                  UNInt64("time"),
                  sample_id())

def thread_map():
    return Struct("thread_map",
                  UNInt64("nr"),
                  Array(lambda ctx: ctx.nr,
                        Struct("thread_map_entry",
                                       UNInt64("pid"),
                                       String("comm", 16))))

def id_index():
    return Struct("id_index",
                  UNInt64("nr"),
                  Array(lambda ctx: ctx.nr,
                     Struct("id_index_entry",
                        UNInt64("id"),
                        UNInt64("idx"),
                        UNInt64("cpu"),
                        UNInt64("tid"))))

def as_is():
    return Embedded(Struct("data", Bytes("data", lambda ctx: ctx.size - 8)))

def throttle(name):
    return Struct(name,
                  UNInt64("time"),
                  UNInt64("id"),
                  UNInt64("stream_id"),
                  sample_id())

def hweight64(ctx):
    return bin(ctx._.attr.perf_event_attr.sample_regs_user).count("1")

def event():
    return Embedded(
        Struct("event",
                If(lambda ctx: sample_type(ctx).identifier,
                   UNInt64("identifier")),
                If(lambda ctx: sample_type(ctx).ip,
                   UNInt64("ip")),
                If(lambda ctx: sample_type(ctx).tid,
                   Embedded(Struct("tid",
                                   SNInt32("pid"),
                                   SNInt32("tid")))),
                If(lambda ctx: sample_type(ctx).time,
                   UNInt64("time")),
                If(lambda ctx: sample_type(ctx).addr,
                   UNInt64("addr")),
                If(lambda ctx: sample_type(ctx).id,
                   UNInt64("id")),
                If(lambda ctx: sample_type(ctx).stream_id,
                   UNInt64("stream_id")),
                If(lambda ctx: sample_type(ctx).cpu,
                   Embedded(Struct("cpu",
                                   UNInt32("cpu"),
                                   UNInt32("res")))),
                If(lambda ctx: sample_type(ctx).period,
                   UNInt64("period")),
                If(lambda ctx: sample_type(ctx).read,
                   read_format()),
                If(lambda ctx: sample_type(ctx).callchain,
                   Struct("callchain",
                          UNInt64("nr"),
                          Array(lambda ctx: ctx.nr,
                                UNInt64("caller")))),
                If(lambda ctx: sample_type(ctx).raw,
                   Struct("raw",
                          UNInt32("size"),
                          Bytes("raw", lambda ctx: ctx.size))),
                If(lambda ctx: sample_type(ctx).branch_stack,
                   Struct("branch_stack",
                          UNInt64("nr"),
                          Array(lambda ctx: ctx.nr,
                                Struct("branch",
                                       UNInt64("from"),
                                       UNInt64("to"),
                                       # Little-Endian!
                                       BitStruct("flags",
                                                 Padding(4),
                                                 Flag("abort"),
                                                 Flag("in_tx"),
                                                 Flag("predicted"),
                                                 Flag("mispred"),
                                                 Padding(64 - 1 * 8)))))),
                If(lambda ctx: sample_type(ctx).regs_user,
                   Struct("regs_user",
                          Enum(UNInt64("abi"),
                               NONE = 0,
                               ABI_32 = 1,
                               ABI_64 = 2),
                          Array(lambda ctx: hweight64(ctx),
                                UNInt64("reg")))),
                If(lambda ctx: sample_type(ctx).stack_user,
                   Struct("stack_user",
                          UNInt64("size"),
                          Bytes("data", lambda ctx: ctx.size),
                          UNInt64("dyn_size"))),
                If(lambda ctx: sample_type(ctx).weight,
                   UNInt64("weight")),
                If(lambda ctx: sample_type(ctx).data_src,
                   UNInt64("data_src")),
                If(lambda ctx: sample_type(ctx).transaction,
                   UNInt64("transaction")),
                If(lambda ctx: sample_type(ctx).regs_intr,
                   Struct("regs_intr",
                          Enum(UNInt64("abi"),
                               NONE = 0,
                               ABI_32 = 1,
                               ABI_64 = 2),
                          Array(lambda ctx: hweight64(ctx),
                                UNInt64("reg")))),
                Anchor("anchor_end_event"),
                Padding(lambda ctx: max(0, ctx.size - ctx.anchor_end_event))))

def get_attr_list(ctx):
    return ctx._._.attrs.perf_file_attr.f_attr

# assume that sample_id_all is the same in all events
# we cannot look up the event without this.
def has_sample_id_all(ctx):
    attr = get_attr_list(ctx)[0]
    if 'sample_id_all' in attr:
        return attr.sample_id_all
    return False

# when sample_id_all is not supported, we may
# not look up the right one (perf.data limitation)
def lookup_event_attr(ctx):
    if "anchor_end_id" in ctx and ctx.anchor_end_id:
        idx = ctx.anchor_end_id
    elif 'id' in ctx and ctx['id']:
        idx = ctx['id']
    else:
        idx = 0
    return get_attr_list(ctx)[idx]

# XXX need to make OnDemand for large files

def perf_event_header():
    return Embedded(Struct(None,
                           Enum(UNInt32("type"),
                                MMAP            = 1,
                                LOST            = 2,
                                COMM            = 3,
                                EXIT            = 4,
                                THROTTLE        = 5,
                                UNTHROTTLE      = 6,
                                FORK            = 7,
                                READ            = 8,
                                SAMPLE          = 9,
                                MMAP2           = 10,
                                RECORD_AUX      = 11,
                                ITRACE_START    = 12,
                                LOST_SAMPLES    = 13,
                                SWITCH          = 14,
                                SWITCH_CPU_WIDE = 15,
                                NAMESPACES      = 16,
                                KSYMBOL         = 17,
                                BPF_EVENT       = 18,
                                CGROUP          = 19,
                                TEXT_POKE       = 20,
                                AUX_OUTPUT_HW_ID= 21,
                                HEADER_ATTR     = 64,
                                HEADER_EVENT_TYPE = 65,
                                TRACING_DATA    = 66,
                                HEADER_BUILD_ID = 67,
                                FINISHED_ROUND  = 68,
                                ID_INDEX        = 69,
                                AUXTRACE_INFO   = 70,
                                AUXTRACE        = 71,
                                AUXTRACE_ERROR  = 72,
                                THREAD_MAP      = 73,
                                CPU_MAP         = 74,
                                STAT_CONFIG     = 75,
                                STAT            = 76,
                                STAT_ROUND      = 77,
                                EVENT_UPDATE    = 78,
                                TIME_CONV       = 79,
                                HEADER_FEATURE  = 80,
                                COMPRESSED      = 81,
                                FINISHED_INIT   = 82),
                           Embedded(BitStruct(None,
                                              Padding(1),
                                              Enum(BitField("cpumode", 7),
                                                   UNKNOWN = 0,
                                                   KERNEL = 1,
                                                   USER = 2,
                                                   HYPERVISOR = 3,
                                                   GUEST_KERNEL = 4,
                                                   GUEST_USER = 5),

                                              Flag("ext_reserved"),
                                              Flag("exact_ip"),
                                              Flag("mmap_data"),
                                              Padding(5))),
                           UNInt16("size"),
                           If(has_sample_id_all,
                                 Pointer(lambda ctx: ctx.anchor_start + ctx.size - 8,
                                   UNInt64("anchor_end_id"))),
                           Value("attr", lookup_event_attr)))

def mmap():
    return Struct("mmap",
                  SNInt32("pid"),
                  SNInt32("tid"),
                  UNInt64("addr"),
                  UNInt64("len"),
                  UNInt64("pgoff"),
                  Anchor("anchor_start_of_filename"),
                  CString("filename"),
                  Anchor("anchor_end_of_filename"),
                  # hack for now. this shouldn't be needed.
                  If(lambda ctx: True, # XXX
                     Embedded(Pointer(lambda ctx:
                                      ctx.size + ctx.anchor_start -
                                      sample_id_size(ctx),
                                      sample_id()))))
def mmap2():
    return Struct("mmap2",
                  SNInt32("pid"),
                  SNInt32("tid"),
                  UNInt64("addr"),
                  UNInt64("len"),
                  UNInt64("pgoff"),
                  UNInt32("maj"),
                  UNInt32("min"),
                  UNInt64("ino"),
                  UNInt64("ino_generation"),
                  UNInt32("prot"),
                  UNInt32("flags"),
                  CString("filename"),
                  sample_id())

def read_flags(ctx):
    return ctx._.attr.read_format

def enabled_running():
    return Struct("enabled_running",
                  If(lambda ctx: read_flags(ctx).total_time_enabled,
                     UNInt64("total_time_enabled")),
                  If(lambda ctx: read_flags(ctx).total_time_running,
                     UNInt64("total_time_running")))

def read_format():
    return Struct("read",
                  If(lambda ctx: read_flags(ctx).group,
                     Struct("group",
                            UNInt64("nr"),
                            Embedded(enabled_running()),
                            Array(lambda ctx: ctx.nr,
                                  Struct("val",
                                         UNInt64("value"),
                                         If(lambda ctx:
                                            read_flags(ctx).id,
                                            UNInt64("id2")))))),
                  If(lambda ctx: not read_flags(ctx).group,
                      Struct("single",
                             UNInt64("value"),
                             Embedded(enabled_running()),
                             If(lambda ctx: read_flags(ctx).id,
                                UNInt64("id2")))))

def time_conv():
    return Struct("time_conv",
                  UNInt64("time_shift"),
                  UNInt64("time_mult"),
                  UNInt64("time_zero"))

def perf_event():
    return Struct("perf_event",
                  Anchor("anchor_start"),
                  perf_event_header(),
                  Switch("data",
                           lambda ctx: ctx.type,
                           {
                              "MMAP": mmap(), # noqa E121
                              "MMAP2": mmap2(),
                              "LOST": Struct("lost",
                                              UNInt64("id"),
                                              UNInt64("lost"),
                                             sample_id()),
                              "COMM": Struct("comm",
                                             SNInt32("pid"),
                                             SNInt32("tid"),
                                             CString("comm"),
                                             sample_id()),
                              "EXIT": fork_exit("exit"),
                              "THROTTLE": throttle("throttle"),
                              "UNTHROTTLE": throttle("unthrottle"),
                              "FINISHED_ROUND": as_is(),
                              "FORK": fork_exit("fork"),
                              "READ": Embedded(Struct("read_event",
                                                      SNInt32("pid"),
                                                      SNInt32("tid"),
                                                      read_format(),
                                                      sample_id())),
                              "SAMPLE": event(),
                              "TIME_CONV": time_conv(),
                              "THREAD_MAP": thread_map(),

                              # below are the so far not handled ones. Dump their
                              # raw data only
                              "RECORD_AUX": as_is(),
                              "AUX": as_is(),
                              "ITRACE_START": as_is(),
                              "LOST_SAMPLES": as_is(),
                              "SWITCH": as_is(),
                              "SWITCH_CPU_WIDE": as_is(),
                              "NAMESPACES": as_is(),
                              "KSYMBOL": as_is(),
                              "BPF_EVENT": as_is(),
                              "CGROUP": as_is(),
                              "TEXT_POKE": as_is(),
                              "AUX_OUTPUT_HW_ID": as_is(),
                              "HEADER_ATTR": as_is(),
                              "HEADER_EVENT_TYPE": as_is(),
                              "TRACING_DATA": as_is(),
                              "HEADER_BUILD_ID": as_is(),
                              "ID_INDEX":id_index(),
                              "AUXTRACE_INFO": as_is(),
                              "AUXTRACE": as_is(),
                              "AUXTRACE_ERROR": as_is(),
                              "CPU_MAP": as_is(),
                              "STAT": as_is(),
                              "STAT_ROUND": as_is(),
                              "EVENT_UPDATE": as_is(),
                              "HEADER_FEATURE": as_is(),
                              "COMPRESSED": as_is(),
                              "FINISHED_INIT": as_is(),
                           }),
                        Anchor("anchor_end"),
                        Padding(lambda ctx:
                                    ctx.size - (ctx.anchor_end - ctx.anchor_start)))

def perf_event_seq(attr):
    return GreedyRange(perf_event(attr))

perf_event_attr_sizes = (64, 72, 80, 96, 104, 112, 120, 128, 136)

perf_event_attr = Struct("perf_event_attr",
                         Anchor("anchor_start"),
                         Enum(UNInt32("type"),
                              HARDWARE = 0,
                              SOFTWARE = 1,
                              TRACEPOINT = 2,
                              HW_CACHE = 3,
                              RAW = 4,
                              BREAKPOINT = 5),
                         UNInt32("size"),
                         UNInt64("config"),
                         UNInt64("sample_period_freq"),
                         # must be in LE order, original is a u64
                         # each byte is reversed
                         BitStruct("sample_type",
                                   Flag("cpu"),
                                   Flag("id"),
                                   Flag("callchain"),
                                   Flag("read"),
                                   Flag("addr"),
                                   Flag("time"),
                                   Flag("tid"),
                                   Flag("ip"),

                                   Flag("data_src"),
                                   Flag("weight"),
                                   Flag("stack_user"),
                                   Flag("regs_user"),
                                   Flag("branch_stack"),
                                   Flag("raw"),
                                   Flag("stream_id"),
                                   Flag("period"),

                                   Padding(5),
                                   Flag("regs_intr"),
                                   Flag("transaction"),
                                   Flag("identifier"),

                                   Padding(64 - 3 * 8)),
                         BitStruct("read_format",
                                   Padding(4),
                                   Flag("group"),
                                   Flag("id"),
                                   Flag("total_time_running"),
                                   Flag("total_time_enabled"),
                                   Padding(64 - 1*8)),
                         Embedded(BitStruct(None,
                                            Flag("disabled"),
                                            Flag("inherit"),
                                            Flag("pinned"),
                                            Flag("exclusive"),
                                            Flag("exclude_user"),
                                            Flag("exclude_kernel"),
                                            Flag("exclude_hv"),
                                            Flag("exclude_idle"),
                                            Flag("mmap"),
                                            Flag("comm"),
                                            Flag("freq"),
                                            Flag("inherit_stat"),
                                            Flag("enable_on_exec"),
                                            Flag("task"),
                                            Flag("watermark"),
                                            BitField("precise_ip", 2),
                                            Flag("mmap_data"),
                                            Flag("sample_id_all"),
                                            Flag("exclude_host"),
                                            Flag("exclude_guest"),
                                            Flag("exclude_callchain_kernel"),
                                            Flag("exclude_callchain_user"),
                                            Padding(41))),
                         UNInt32("wakeup_events"),
                         UNInt32("bp_type"),
                         UNInt64("config1"),
                         If(lambda ctx: ctx.size >= perf_event_attr_sizes[1],
                            UNInt64("config2")),
                         If(lambda ctx: ctx.size >= perf_event_attr_sizes[2],
                            UNInt64("branch_sample_type")),
                         If(lambda ctx: ctx.size >= perf_event_attr_sizes[3],
                            Embedded(Struct(None,
                                            UNInt64("sample_regs_user"),
                                            UNInt32("sample_stack_user"),
                                            UNInt32("__reserved_2")))),
                         If(lambda ctx: ctx.size >= perf_event_attr_sizes[4],
                            UNInt64("sample_regs_intr")),
                         If(lambda ctx: ctx.size >= perf_event_attr_sizes[5],
                            UNInt32("aux_watermark")),
                         If(lambda ctx: ctx.size >= perf_event_attr_sizes[6],
                            UNInt32("aux_sample_size")),
                         If(lambda ctx: ctx.size >= perf_event_attr_sizes[7],
                            UNInt64("sig_data")),
                         If(lambda ctx: ctx.size >= perf_event_attr_sizes[8],
                            UNInt64("config3")),
                         Anchor("anchor_end"),
                         Padding(lambda ctx: ctx.size - ctx.anchor_end + ctx.anchor_start))

def pad(l = "len"):
    return Padding(lambda ctx: ctx[l] - (ctx.anchor_offset - ctx.anchor_start))

def str_with_len(name):
    return Struct(name,
                  UNInt32("len"),
                  Anchor("anchor_start"),
                  CString(name),
                  Anchor("anchor_offset"),
                  pad())

def feature_string(name):
    return If(lambda ctx: ctx._[name],
              perf_file_section(name,
                                Embedded(Struct(name,
                                                UNInt32("len"),
                                                CString(name)))))

def string_list(name, extra = Pass):
    return PrefixedArray(Struct(name,
                               UNInt32("len"),
                               Anchor("anchor_start"),
                               CString(name),
                               Anchor("anchor_offset"),
                               pad(),
                               extra), UNInt32("nr"))

def numa_topology():
    return PrefixedArray(Struct("node",
                                UNInt32("nodenr"),
                                UNInt64("mem_total"),
                                UNInt64("mem_free"),
                                str_with_len("cpus")),
                         UNInt32("nr"))

def group_desc():
    return string_list("group_desc",
                       Embedded(Struct(None,
                                       UNInt32("leader_idx"),
                                       UNInt32("nr_members"))))

def feature_bytes(name):
    return Struct(name,
                  UNInt64("offset"),
                  UNInt64("size"),
                  Pointer(lambda ctx: ctx.offset, Bytes("data", lambda ctx: ctx.size)))

def build_id():
    return Struct("build_id",
                  Anchor("anchor_start"),
                  UNInt32("type"),
                  UNInt16("misc"),
                  UNInt16("size"),
                  SNInt32("pid"),
                  HexDumpAdapter(String("build_id", 24)),
                  CString("filename"),
                  Anchor("anchor_offset"),
                  pad("size"))

def section_adapter(name, target):
    return perf_file_section(name,
                             TunnelAdapter(String("data",
                                                  lambda ctx: ctx.size),
                                           target))

def pmu_mappings():
    return PrefixedArray(Struct("pmu",
                                UNInt32("type"),
                                str_with_len("name")),
                         UNInt32("nr"))

def event_desc():
    return Struct("event_desc",
                  UNInt32("nr"),
                  UNInt32("attr_size"),
                  Array(lambda ctx: ctx.nr,
                        Struct("desc",
                               perf_event_attr,
                               UNInt32("nr_ids"),
                               str_with_len("event"),
                               Array(lambda ctx: ctx.nr_ids,
                                     UNInt64("id")))))

def perf_features():
    return Struct("features",
                  # XXX
                  If(lambda ctx: ctx._.tracing_data,
                     perf_file_section("tracing_data",
                                       Pass)),
                  If(lambda ctx: ctx._.build_id,
                     section_adapter("build_id",
                                     GreedyRange(build_id()))),
                  feature_string("hostname"),
                  feature_string("osrelease"),
                  feature_string("version"),
                  feature_string("arch"),
                  If(lambda ctx: ctx._.nrcpus,
                     perf_file_section("nrcpus",
                                       Embedded(Struct("nrcpus",
                                                UNInt32("nr_cpus_online"),
                                                UNInt32("nr_cpus_avail"))))),
                  feature_string("cpudesc"),
                  feature_string("cpuid"),
                  If(lambda ctx: ctx._.total_mem,
                     perf_file_section("total_mem",
                                       UNInt64("total_mem"))),
                  If(lambda ctx: ctx._.cmdline,
                     perf_file_section("cmdline",
                                       string_list("cmdline"))),
                  If(lambda ctx: ctx._.event_desc,
                     perf_file_section("event_desc",
                                       event_desc())),
                  If(lambda ctx: ctx._.cpu_topology,
                     perf_file_section("cpu_topology",
                                       Struct("cpu_topology",
                                              string_list("cores"),
                                              string_list("threads")))),
                  If(lambda ctx: ctx._.numa_topology,
                     perf_file_section("numa_topology",
                                       numa_topology())),
                  # not implemented in perf
                  If(lambda ctx: ctx._.branch_stack,
                     perf_file_section("branch_stack",
                                       Pass)),
                  If(lambda ctx: ctx._.pmu_mappings,
                     perf_file_section("pmu_mappings",
                                       pmu_mappings())),
                  If(lambda ctx: ctx._.group_desc,
                     perf_file_section("group_desc",
                                       group_desc())),
                  If(lambda ctx: ctx._.auxtrace,
                     feature_bytes("auxtrace")),
                  If(lambda ctx: ctx._.stat,
                     feature_bytes("stat")),
                  If(lambda ctx: ctx._.cache,
                     feature_bytes("cache")),
                  If(lambda ctx: ctx._.sample_time,
                     feature_bytes("sample_time")),
                  If(lambda ctx: ctx._.mem_tpology,
                     feature_bytes("mem_tpology")),
                  If(lambda ctx: ctx._.clock_id,
                     feature_bytes("clock_id")),
                  If(lambda ctx: ctx._.dir_format,
                     feature_bytes("dir_format")),
                  If(lambda ctx: ctx._.bpf_prog_info,
                     feature_bytes("bpf_prog_info")),
                  If(lambda ctx: ctx._.bpf_btf,
                     feature_bytes("bpf_btf")),
                  If(lambda ctx: ctx._.compressed,
                     feature_bytes("compressed")),
                  If(lambda ctx: ctx._.pmu_caps,
                     feature_bytes("pmu_caps")),
                  If(lambda ctx: ctx._.clock_data,
                     feature_bytes("clock_data")),
                  If(lambda ctx: ctx._.hybrid_topology,
                     feature_bytes("hybrid_topology")),
                  If(lambda ctx: ctx._.cpu_pmu_caps,
                     feature_bytes("cpu_pmu_caps")),)

def perf_file_section(name, target):
    return Struct(name,
                  UNInt64("offset"),
                  UNInt64("size"),
                  Pointer(lambda ctx: ctx.offset, target))

id_array = Array(lambda ctx: ctx.size / 8,
                 UNInt64("id"))

def num_attr(ctx):
    return ctx._.size / ctx._._.attr_size

perf_file_attr = Struct("perf_file_attr",
                        Array(lambda ctx: num_attr(ctx),
                              Struct("f_attr",
                                     perf_event_attr,
                                     perf_file_section("ids", id_array))))

perf_event_types = Struct("perf_file_attr",
                          Anchor("anchor_here"),
                          Padding(lambda ctx: ctx._.size))

perf_data = TunnelAdapter(Bytes("perf_data", lambda ctx: ctx.size),
                          GreedyRange(perf_event()))

#OnDemand(Bytes("perf_data", lambda ctx: ctx.size))

perf_file = Struct("perf_file_header",
                   # no support for version 1
                   Magic("PERFILE2"),
                   UNInt64("size"),
                   UNInt64("attr_size"),
                   perf_file_section("attrs", perf_file_attr),
                   perf_file_section("data", perf_data),
                   perf_file_section("event_types", perf_event_types),
                   # little endian
                   Embedded(BitStruct(None,
                             Flag("nrcpus"),
                             Flag("arch"),
                             Flag("version"),
                             Flag("osrelease"),
                             Flag("hostname"),
                             Flag("build_id"),
                             Flag("tracing_data"),
                             Flag("reserved"),

                             Flag("branch_stack"),
                             Flag("numa_topology"),
                             Flag("cpu_topology"),
                             Flag("event_desc"),
                             Flag("cmdline"),
                             Flag("total_mem"),
                             Flag("cpuid"),
                             Flag("cpudesc"),

                             Flag("clock_id"),
                             Flag("mem_tpology"),
                             Flag("sample_time"),
                             Flag("cache"),
                             Flag("stat"),
                             Flag("auxtrace"),
                             Flag("group_desc"),
                             Flag("pmu_mappings"),

                             Flag("pmu_caps"),
                             Flag("hybrid_topology"),
                             Flag("clock_data"),
                             Flag("cpu_pmu_caps"),
                             Flag("compressed"),
                             Flag("bpf_btf"),
                             Flag("bpf_prog_info"),
                             Flag("dir_format"),

                             Padding(256 - 4*8))),
                   Pointer(lambda ctx: ctx.data.offset + ctx.data.size,
                           perf_features()),)

def get_events(h):
    return h.data.perf_data

if __name__ == '__main__':
    import argparse

    args = argparse.ArgumentParser()
    args.add_argument('file', help='perf.data to read', default='perf.data',
                      nargs='?')
    p = args.parse_args()

    with open(p.file, "rb") as f:
        h = perf_file.parse_stream(f)
        print(h)
        #print(get_events(h))



#!/usr/bin/python
# extract utilized CPUs out of toplev CSV output
# toplev ... -I 1000 --node +CPU_Utilization -x, -o x.csv ...
# utilized.py < x.csv
# note it duplicates the core output
from __future__ import print_function
import argparse
import csv
import sys
import re
import collections

ap = argparse.ArgumentParser()
ap.add_argument('--min-util', default=10., type=float)
ap.add_argument('file', nargs='?', type=argparse.FileType('r'), default=sys.stdin)
ap.add_argument('--output', '-o', type=argparse.FileType('w'), default=sys.stdout)
args = ap.parse_args()

key = None

c = csv.reader(args.file)
wr = csv.writer(args.output)

fields = collections.OrderedDict()
util = collections.defaultdict(list)

for t in c:
    if len(t) < 3 or t[0].startswith("#"):
        continue
    if t[0] == "Timestamp":
        wr.writerow(t)
    key = t[1] # XXX handle no -I
    if key in fields:
        fields[key].append(t)
    else:
        fields[key] = [t]
    if t[2] == "CPU_Utilization":
        util[key].append(float(t[3]))

final = []
skipped = []
for j in fields.keys():
    if "-T" not in j and not j.startswith("CPU"):
        if "S" in j:
            final.append(j)
        continue
    core = re.sub(r'-T\d+', '', j)
    utilization = 100
    if len(util[j]) > 0:
        utilization = (sum(util[j]) / len(util[j])) * 100.
    if utilization >= float(args.min_util):
        for k in fields[core] + fields[j]:
            wr.writerow(k)
    else:
        skipped.append(j)
for j in final:
    for k in fields[j]:
        wr.writerow(k)
print("skipped", " ".join(skipped), file=sys.stderr)



#!/usr/bin/env python3
# print currently running events on cpu (default 0)
# event-rmap [cpu-num]
# xxx no extra modi for now, racy with multi plexing
from __future__ import print_function
import sys
import msr
import ocperf
from pmudef import (MSR_PEBS_ENABLE, MSR_EVNTSEL, EVENTSEL_ENABLE, EVMASK,
                    EVENTSEL_CMASK,
                    EVENTSEL_EDGE, EVENTSEL_ANY, EVENTSEL_INV, EVENTSEL_PC,
                    MSR_IA32_FIXED_CTR_CTRL)

fixednames = (
    "inst_retired.any",
    "cpu_clk_unhalted.thread",
    "cpu_clk_unhalted.ref_tsc"
)

cpu = 0
if len(sys.argv) > 1:
    cpu = int(sys.argv[1])

emap = ocperf.find_emap()
if not emap:
    print("Unknown CPU or cannot find CPU event table")
found = 0
try:
    pebs_enable = msr.readmsr(MSR_PEBS_ENABLE, cpu)
except OSError:
    pebs_enable = 0
for i in range(0, 8):
    try:
        evsel = msr.readmsr(MSR_EVNTSEL + i, cpu)
    except OSError:
        break
    found += 1
    if evsel & EVENTSEL_ENABLE:
        print("%d: %016x: " % (i, evsel), end="")
        evsel &= EVMASK
        if emap is None:
            name = "r%04x", evsel & 0xffff
        elif evsel in emap.codes:
            ev = emap.codes[evsel]
            if ev.msr:
                try:
                    extra = msr.readmsr(ev.msr)
                except OSError:
                    print("Cannot read extra MSR %x for %s" % (ev.msr, ev.name))
                    continue
                for j in emap.codes.keys():
                    if j == evsel and extra == emap.codes[j].msrvalue:
                        print(j.name, "msr:%x" % (extra), end="")
                        break
                else:
                    print("no exact match for %s, msr %x value %x" % (ev.name,
                                                                      ev.msr, ev.msrvalue), end="")
            else:
                print(ev.name, end="")
        else:
            name = ""
            for j in emap.codes.keys():
                if j & 0xff == evsel & 0xff:
                    name += "%s[%x] " % (emap.codes[j].name, j)
            if name:
                print("[no exact match] " + name, end=" ")
            else:
                print("r%x" % (evsel), end=" ")
        if evsel & EVENTSEL_CMASK:
            print("cmask=%x" % (evsel >> 24), end=" ")
        if evsel & EVENTSEL_EDGE:
            print("edge=1", end=" ")
        if evsel & EVENTSEL_ANY:
            print("any=1", end=" ")
        if evsel & EVENTSEL_INV:
            print("inv=1", end=" ")
        if evsel & EVENTSEL_PC:
            print("pc=1", end=" ")
        if pebs_enable & (1 << i):
            print("precise=1", end=" ")
        print()
if found == 0:
    print("Cannot read any MSRs")

try:
    fixed = msr.readmsr(MSR_IA32_FIXED_CTR_CTRL)
except OSError:
    print("Cannot read fixed counter MSR")
    fixed = 0
for i in range(0, 2):
    if fixed & (1 << (i*4)):
        print("fixed %d: %s" % (i, fixednames[i]))



#!/usr/bin/env python
# Import perf.data into a pandas DataFrame
#
# Copyright (c) 2013-2014, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
from __future__ import print_function
import pandas as pd
import numpy as np
import perfdata
from collections import defaultdict, Counter
import elf
import mmap

ignored = {'type', 'start', 'end', '__recursion_lock__', 'ext_reserved',
           'header_end', 'end_event', 'offset', 'callchain', 'branch',
           'branch_stack', 'end_id', 'size', 'cpumode', 'caller', 'time',
           # skip attr for now, as it is too complex
           # XXX simple representation
           'attr'}

bool_fields = {'kernel', 'hv', 'guest'}

def resolve_list(j, ip, mm, need_line):
    filename, _, foffset = mm.resolve(j.pid, ip)
    sym, soffset, line = elf.resolve_ip(filename, foffset, ip, need_line)
    return [filename, sym, soffset, line]

def resolve_chain(cc, j, mm, need_line):
    if not cc:
        return []
    res = []
    for ip in cc.caller:
        r = [ip,]
        r += resolve_list(j, ip, mm, need_line)
        res.append(r)
    return res

def resolve_branch(branch, j, mm, need_line):
    res = []
    for br in branch:
        # XXX flags
        r = [br['from'], br['to']]
        r += resolve_list(j, br['from'], mm, need_line)
        r += resolve_list(j, br['to'], mm, need_line)
        res.append(r)
    return res

class Path:
    """Store either a callchain or a branch stack as a list with id."""
    def __init__(self, val, id):
        self.val = val
        self.id = id

class Aux:
    """Store auxilliary data to the main pandas perf array, like call chains
       or branch stacks. The data is deduped and a unique id generated."""

    def __init__(self):
        self.ids = dict()
        self.paths = dict()
        self.next_id = 0

    def alloc_id(self):
        id = self.next_id
        self.next_id += 1
        return id

    def add(self, h, create):
        h = tuple(h)
        if h in self.paths:
            return self.paths[h].id
        id = self.alloc_id()
        path = Path(create(), id)
        self.paths[h] = path
        self.ids[id] = path
        return id

    def getid(self, id):
        return self.ids[id]

    def __getitem__(self, id):
        return self.ids[id]

cpumodes = {
    'UNKNOWN': (0, 0, 0),
    'KERNEL': (1, 0, 0),
    'USER': (0, 0, 0),
    'HYPERVISOR': (0, 1, 0),
    'GUEST_KERNEL': (1, 0, 1),
    'GUEST_USER': (0, 0, 1),
}

def samples_to_df(h, need_line):
    """Convert a parsed perf event list to a pandas table.
       The pandas table contains all events in a easy to process format.
       The pandas table has callchain_aux and branch_aux fields pointing
       to Aux object defining the callchains/branches."""
    ev = perfdata.get_events(h)
    index = []
    data = defaultdict(list)
    callchains = Aux()
    branches = Aux()

    used = Counter()
    mm = mmap.MmapTracker()

    numsample = 0
    for n in range(0, len(ev)):
        mm.lookahead_mmap(ev, n)

        j = ev[n]
        print(j)
        if j.type != "SAMPLE":
            continue

        numsample += 1

        mm.update_sample(j)
        def add(k, i):
            data[k].append(i)
            used[k] += 1

        filename, mmap_base, foffset = mm.resolve(j.pid, j.ip)
        if filename == "[kernel.kallsyms]_text":
            filename = None
        add('filename', filename)
        sym, soffset, line = elf.resolve_ip(filename, foffset, j.ip, need_line)
        add('symbol', sym)
        add('line', line)
        add('soffset', soffset)
        if 'callchain' in j and j.callchain:
            id = callchains.add(j.callchain.caller,
                    lambda: resolve_chain(j.callchain, j, mm, need_line))
            add('callchain', id)
        if 'branch_stack' in j and j.branch_stack:
            branch = j.branch_stack.branch
            id = branches.add(map(lambda x: (x['from'], x.to), branch),
                    lambda: resolve_branch(branch, j, mm, need_line))
            add('branch', id)
        kernel, guest, hv = cpumodes[j['cpumode']]
        add('kernel', kernel)
        add('guest', guest)
        add('hv', hv)
        for name in j:
            if name not in ignored:
                if j[name]:
                    used[name] += 1
                data[name].append(j[name])
        index.append(int(j["time"]))
    for j in data.keys():
        if used[j] == 0:
            del data[j]
    df = pd.DataFrame(data, index=index, dtype=np.uint64)
    if numsample > 0:
        for i in bool_fields:
            df[i] = df[i].astype('bool')
    df.branch_aux = branches
    df.callchain_aux = callchains
    return df

def read_samples(fn, need_line=True):
    with open(fn, "rb") as f:
        h = perfdata.perf_file.parse_stream(f)
        df = samples_to_df(h, need_line)
        return df, h.attrs.perf_file_attr.f_attr, h.features

if __name__ == '__main__':
    import argparse
    import sys

    args = argparse.ArgumentParser()
    args.add_argument('file', nargs='?', help='perf.data file to read',
                          default='perf.data')
    args.add_argument('--repl', action='store_true',
                             help='start python shell with data')
    args.add_argument('--ipython', action='store_true',
                             help='start ipython shell with data')
    p = args.parse_args()
    df, _, _ = read_samples(p.file)
    if p.repl:
        import code
        print(df)
        code.interact(banner='perf.data is in df', local=locals())
        sys.exit(0)

    if p.ipython:
        try:
            from IPython.terminal.embed import InteractiveShellEmbed
        except NameError:
            sys.exit("Ipython not installed")
        print(df)
        ipshell = InteractiveShellEmbed(banner1="perf.data is in df")
        ipshell()
        sys.exit(0)

    print(df)
    if 'filename' in df:
        print(df['filename'].value_counts())
    if 'symbol' in df:
        print(df['symbol'].value_counts())
    if 'line' in df:
        print(df['line'].value_counts())



#!/usr/bin/env python3
# Copyright (c) 2014-2020, Intel Corporation
# Author: Andi Kleen
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# Automatic event list downloader
#
# event_download.py         download for current cpu
# event_download.py -a      download all
# event_download.py cpustr...  Download for specific CPU
#
# env:
# CPUINFO=... override /proc/cpuinfo file
# MAPFILE=... override mapfile.csv
# PERFMONDIR=... override download prefix for perfmon data, can be a local clone (file:///tmp/perfmon)
from __future__ import print_function
import sys
import re
try:
    from urllib.request import urlopen
    from urllib.error import URLError
except ImportError:
    from urllib2 import urlopen, URLError
import os
import string
from fnmatch import fnmatch

urlpath = os.environ.get('PERFMONDIR', 'https://raw.githubusercontent.com/intel/perfmon/main')
mapfile = 'mapfile.csv'
modelpath = urlpath + "/" + mapfile

def get_cpustr():
    cpuinfo = os.getenv("CPUINFO")
    if cpuinfo is None:
        cpuinfo = '/proc/cpuinfo'
    f = open(cpuinfo, 'r')
    cpu = [None, None, None, None]
    for j in f:
        n = j.split()
        if n[0] == 'vendor_id':
            cpu[0] = n[2]
        elif n[0] == 'model' and n[1] == ':':
            cpu[2] = int(n[2])
        elif n[0] == 'cpu' and n[1] == 'family':
            cpu[1] = int(n[3])
        elif n[0] == 'stepping' and n[1] == ':':
            cpu[3] = int(n[2])
        if all(v is not None for v in cpu):
            break
    # stepping for SKX only
    stepping = cpu[0] == "GenuineIntel" and cpu[1] == 6 and cpu[2] == 0x55
    if stepping:
        return "%s-%d-%X-%X" % tuple(cpu)
    return "%s-%d-%X" % tuple(cpu)[:3]

def sanitize(s, a):
    o = ""
    for j in s:
        if j in a:
            o += j
    return o

def getdir():
    try:
        d = os.getenv("XDG_CACHE_HOME")
        xd = d
        if not d:
            home = os.getenv("HOME")
            d = "%s/.cache" % (home)
        d += "/pmu-events"
        if not os.path.isdir(d):
            # try to handle the sudo case
            if not xd:
                user = os.getenv("SUDO_USER")
                if user:
                    nd = os.path.expanduser("~" + user) + "/.cache/pmu-events"
                    if os.path.isdir(nd):
                        return nd
            os.makedirs(d)
        return d
    except OSError:
        raise Exception('Cannot access ' + d)

NUM_TRIES = 3

def getfile(url, dirfn, fn):
    tries = 0
    print("Downloading", url, "to", fn)
    while True:
        try:
            f = urlopen(url)
            data = f.read()
        except IOError:
            tries += 1
            if tries >= NUM_TRIES:
                raise
            print("retrying download")
            continue
        break
    o = open(os.path.join(dirfn, fn), "wb")
    o.write(data)
    o.close()
    f.close()

printed = set()

def warn_once(msg):
    if msg not in printed:
        print(msg, file=sys.stderr)
        printed.add(msg)

def cpu_without_step(match):
    if match.count("-") < 3:
        return match
    n = match.split("-")
    return "%s-%s-%s" % tuple(n[:3])

allowed_chars = string.ascii_letters + '_-.' + string.digits
def parse_map_file(match, key=None, link=True, onlyprint=False,
                   acceptfile=False, hybridkey=None):
    match2 = cpu_without_step(match)
    files = []
    dirfn = getdir()
    try:
        mfn = os.getenv("MAPFILE")
        if mfn:
            mapfn = mfn
            acceptfile = True
        else:
            mapfn = os.path.join(dirfn, mapfile)
        if onlyprint and not os.path.exists(mapfn) and not mfn:
            print("Download", mapfn, "first for --print")
            return []
        if acceptfile and os.path.exists(mapfn):
            pass
        elif not onlyprint and not mfn:
            getfile(modelpath, dirfn, mapfile)
        models = open(mapfn)
        for j in models:
            if j.startswith("Family-model"):
                continue
            n = j.rstrip().split(",")
            if len(n) < 4:
                if len(n) > 0:
                    print("Cannot parse", n)
                continue
            cpu, version, name, typ = n[:4]
            if not (fnmatch(cpu, match) or fnmatch(cpu, match2) or
                    fnmatch(match2, cpu) or fnmatch(match, cpu)):
                continue
            if key is not None and typ != key:
                continue
            hkey = n[6] if len(n) >= 7 else None
            if hybridkey and hkey and hkey != hybridkey:
                continue
            cpu = sanitize(cpu, allowed_chars)
            url = urlpath + name
            matchfn = match
            if matchfn == "*":
                matchfn = cpu
            if ".json" not in matchfn:
                if hkey:
                    fn = "%s-%s-%s.json" % (matchfn, sanitize(typ, allowed_chars), hkey)
                else:
                    fn = "%s-%s.json" % (matchfn, sanitize(typ, allowed_chars))
            path = os.path.join(dirfn, fn)
            if acceptfile and os.path.exists(path):
                if onlyprint:
                    print(path)
                    continue
            else:
                try:
                    os.remove(path)
                except OSError:
                    pass
                if onlyprint:
                    print(path)
                    continue
                if mfn:
                    print("error accessing", path)
                    continue
                try:
                    fn = fn.replace("01234", "4")
                    fn = fn.replace("56789ABCDEF", "5") # XXX
                    getfile(url, dirfn, fn)
                except URLError as e:
                    print("error accessing %s: %s" % (url, e), file=sys.stderr)
                    if match == '*':
                        continue
                    raise
            if link:
                lname = re.sub(r'.*/', '', name)
                lname = sanitize(lname, allowed_chars)
                try:
                    os.remove(os.path.join(dirfn, lname))
                except OSError:
                    pass
                try:
                    os.symlink(fn, os.path.join(dirfn, lname))
                except OSError as e:
                    print("Cannot link %s to %s:" % (name, lname), e, file=sys.stderr)
            files.append(fn)
        models.close()
        for file_name in ["README.md", "LICENSE"]:
            if not onlyprint and not os.path.exists(os.path.join(dirfn, file_name)) and not mfn:
                getfile(urlpath + "/" + file_name, dirfn, file_name)
    except URLError as e:
        print("Cannot access event server:", e, file=sys.stderr)
        warn_once("""
If you need a proxy to access the internet please set it with:
\texport https_proxy=http://proxyname...
If you are not connected to the internet please run this on a connected system:
\tevent_download.py '%s'
and then copy ~/.cache/pmu-events to the system under test
To get events for all possible CPUs use:
\tevent_download.py -a""" % match)
    except OSError as e:
        print("Cannot write events file:", e, file=sys.stderr)
    return files

def download(match, key=None, link=True, onlyprint=False, acceptfile=False):
    return len(parse_map_file(match, key, link, onlyprint, acceptfile))

def download_current(link=False, onlyprint=False):
    """Download JSON event list for current cpu.
       Returns >0 when a event list is found"""
    return download(get_cpustr(), link=link, onlyprint=onlyprint, )

def qualify_fn(cache, fn):
    if os.path.exists(fn):
        return fn
    nfn = "%s/%s" % (cache, fn)
    if os.path.exists(nfn):
        return nfn
    return fn

def eventlist_name(name=None, key="core", hybridkey=None):
    if not name:
        name = get_cpustr()
    cache = getdir()
    fn = name
    if os.path.exists(fn):
        return fn
    if ".json" not in name:
        if hybridkey:
            fn = "%s-%s-%s.json" % (name, key, hybridkey)
        else:
            fn = "%s-%s.json" % (name, key)
    if "/" not in fn:
        fn = "%s/%s" % (cache, fn)
    if not os.path.exists(fn):
        files = parse_map_file(name, key, acceptfile=True, hybridkey=hybridkey)
        if files:
            return qualify_fn(cache, files[0])
        name = cpu_without_step(name)
        if "*" in fn:
            fn = "%s/%s" % (cache, name)
        elif hybridkey:
            fn = "%s/%s-%s-%s.json" % (cache, name, key, hybridkey)
        else:
            fn = "%s/%s-%s.json" % (cache, name, key)
        files = parse_map_file(name, key, acceptfile=True, hybridkey=hybridkey)
        if files:
            fn = files[0]
    return qualify_fn(cache, fn)

if __name__ == '__main__':
    # only import argparse when actually called from command line
    # this makes ocperf work on older python versions without it.
    import argparse
    p = argparse.ArgumentParser(usage='download Intel event files')
    p.add_argument('--all', '-a', help='Download all available event files', action='store_true')
    p.add_argument('--verbose', '-v', help='Be verbose', action='store_true')
    p.add_argument('--mine', help='Print name of current CPU', action='store_true')
    p.add_argument('--link', help='Create links with the original event file name', action='store_true', default=True)
    p.add_argument('--print', help='Print file names of all event files instead of downloading. Requires existing mapfile.csv.',
                   dest='print_', action='store_true')
    p.add_argument('cpus', help='CPU identifiers to download', nargs='*')
    args = p.parse_args()

    if args.verbose or args.mine:
        print(get_cpustr())
    if args.mine:
        sys.exit(0)
    d = getdir()
    if args.all:
        found = download('*', link=args.link, onlyprint=args.print_)
    elif len(args.cpus) == 0:
        found = download_current(link=args.link, onlyprint=args.print_)
    else:
        found = 0
        for j in args.cpus:
            found += download(j, link=args.link, onlyprint=args.print_)

    if found == 0 and not args.print_:
        print("Nothing found", file=sys.stderr)

    el = eventlist_name()
    if os.path.exists(el) and not args.print_:
        print("my event list", el)



# Helper classes and functions for nodes

# Decorator class to declare reference dependecies between classes
class requires(object):
    """Decorator to mark required references. These references will
    be added to the object as instance attributes. Example:

    @requires("ref1", "ref2")
    class SomeClass(object):
        def some_method(self):
            return self.ref1 + self.ref2

    """
    def __init__(self, *required_refs):
        self.required_refs = required_refs

    def __call__(self, cls):
        setattr(cls, "required_refs", self.required_refs)
        return cls

def set_parent(parent, nodes):
    for node in nodes:
        node.parent = parent

# Check that all required references are set
def check_refs(fn):
    """Decorator to check if required references for an object
    are set. If it finds missing references, it will raise an
    exception. Example:

    @requires("retiring", "bad_speculation", "frontend_bound")
    class BackendBound(object):
        @check_refs
        def _compute(self, ev):
            # checks if required refs are set before executing

    """
    def wrapped(self, *args, **kwargs):
        if not hasattr(self, "required_refs"):
            raise Exception("Missing required_refs object")
        missing_refs = [ref for ref in self.required_refs
                        if not hasattr(self, ref)]
        if missing_refs:
            raise Exception("Missing references: {0}".format(missing_refs))

        return fn(self, *args, **kwargs)

    wrapped.__name__ = fn.__name__
    return wrapped

def add_references(node, **refs):
    """Adds an attribute to node, as specified in the **refs argument.
    Example:

    ...
    backend = BackendBound()
    add_references(backend, retiring=retiring, frontend_bound=frontend,
                   bad_speculation=bad_speculation)

    """
    for name, obj in refs.items():
        setattr(node, name, obj)



#!/usr/bin/env python3
# translate raw events to names
# event-translate rXXX ...
from __future__ import print_function
import re
import sys
import ocperf
from pmudef import EVMASK, EVENTSEL_EVENT, EVENTSEL_UMASK, extra_flags

emap = ocperf.find_emap()
if not emap:
    sys.exit("Unknown CPU or cannot find event table")
for j in sys.argv[1:]:
    m = re.match(r'r([0-9a-f]+)(:.*)?', j)
    if m:
        print(m.group(1))
        evsel = int(m.group(1), 16)
        print("%s:" % (j))
        if evsel & EVMASK in emap.codes:
            print(emap.codes[evsel & EVMASK].name)
        elif (evsel & (EVENTSEL_EVENT|EVENTSEL_UMASK)) in emap.codes:
            print(emap.codes[evsel & (EVENTSEL_EVENT|EVENTSEL_UMASK)].name, end='')
            for k in extra_flags:
                if evsel & k[0]:
                    m = k[0]
                    en = evsel
                    while (m & 1) == 0:
                        m >>= 1
                        en >>= 1
                    print("%s=%d" % (k[1], en & m),end='')
            print()
        else:
            print("cannot find", m.group(1))
    else:
        # XXX implement offcore new style events
        print("cannot parse", j)



#!/usr/bin/env python3
# library and tool to access PCI config space
import os
import struct

# no multiple domains, controllers so far

def probe(bus, dev, func):
    fn = "/sys/devices/pci0000:%02x/0000:%02x:%02x.%01x/config" % (bus, bus, dev, func)
    return os.path.isfile(fn)

def openpci(bus, dev, func, offset, mode):
    fn = "/sys/devices/pci0000:%02x/0000:%02x:%02x.%01x/config" % (bus, bus, dev, func)
    f = os.open(fn, mode)
    os.lseek(f, offset, os.SEEK_SET)
    return f

sizes = {8: "Q", 4: "I", 2: "H", 1: "B"}

def writepci(bus, device, func, offset, size, val):
    f = openpci(bus, device, func, offset, os.O_WRONLY)
    os.write(f, struct.pack(sizes[size], val))
    os.close(f)

def readpci(bus, device, func, offset, size):
    f = openpci(bus, device, func, offset, os.O_RDONLY)
    v = struct.unpack(sizes[size], os.read(f, size))[0]
    os.close(f)
    return v

def changebit(bus, device, func, offset, bit, val):
    f = openpci(bus, device, func, offset, os.O_RDWR)
    v = struct.unpack("I", os.read(f, 4))[0]
    if val:
        v = v | (1 << bit)
    else:
        v = v & ~(1 << bit)
    os.lseek(f, offset, os.SEEK_SET)
    os.write(f, struct.pack('I', v))
    os.close(f)
